import 'dart:convert';
import 'dart:typed_data';
import 'package:http/http.dart' as http;
import 'package:modular_system_core/mscore.dart';

enum HttpMethod { get, post, put, delete, options }
enum DataFormat { json, xml, html }

abstract class Http {
  Future<HttpOutput> sendAsync(HttpInput request);
}

class ConcreteHttp implements Http {
  @override
  Future<HttpOutput> sendAsync(HttpInput request) {
    var method = request.method;
    if (method == HttpMethod.get) {
      return getAsync(request);
    } else if (method == HttpMethod.post) {
      return postAsync(request);
    } else if (method == HttpMethod.put) {
      return putAsync(request);
    } else if (method == HttpMethod.delete) {
      return deleteAsync(request);
    } else {
      throw AppException("Invalid http method", ExceptionCode.internal, StackTrace.current);
    }
  }

  Future<HttpOutput> getAsync(HttpInput request) async {
    var resp = await http.get(request.uri);
    return parse(resp);
  }

  Future<HttpOutput> postAsync(HttpInput request) async {
    var resp = await http.post(
      request.uri,
      headers: request.headers,
      encoding: request.body?.encoding,
      body: request.body?.convert(),
    );
    return parse(resp);
  }

  Future<HttpOutput> putAsync(HttpInput request) async {
    var resp = await http.put(
      request.uri,
      headers: request.headers,
      encoding: request.body?.encoding,
      body: request.body?.convert(),
    );
    return parse(resp);
  }

  Future<HttpOutput> deleteAsync(HttpInput request) async {
    var resp = await http.delete(
      request.uri,
      headers: request.headers,
      encoding: request.body?.encoding,
      body: request.body?.convert(),
    );
    return parse(resp);
  }

  HttpOutput parse(http.Response resp) {
    var output = HttpOutput(isOk: resp.statusCode == 200, statusCode: resp.statusCode);
    output.statusCode = resp.statusCode;
    output.message = getMessageFrom(resp);
    output.encoding = getEncodingFrom(resp);
    output.data = getDataFrom(resp);
    return output;
  }

  DataFormat? getDataFormatFrom(http.Response resp) {
    var type = resp.headers["content-type"]?.toLowerCase();
    if (type == null) return null;
    if (type.contains("application/json")) return DataFormat.json;
    if (type.contains("application/xml")) return DataFormat.xml;
    if (type.contains("text/html")) return DataFormat.html;
    return null;
  }

  String? getMessageFrom(http.Response resp) {
    return resp.reasonPhrase;
  }

  Encoding? getEncodingFrom(http.Response resp) {
    return utf8;
  }

  Uint8List? getDataFrom(http.Response resp) {
    return resp.bodyBytes;
  }
}

class HttpInput {
  HttpInput({
    required this.uri,
    required this.method,
    this.headers,
    this.body,
  }) {
    setHeaders();
  }

  Uri uri;
  HttpMethod method;
  Map<String, String>? headers;
  HttpBody? body;

  void setHeaders() {
    headers ??= {};
    setContentTypeHeader();
  }

  void setContentTypeHeader() {
    if (body?.format == DataFormat.json) {
      headers?.add("content-type", "${getMimeType()}; ${getEncoding()}");
    }
  }

  String getMimeType() {
    if (body?.format == DataFormat.json) {
      return "application/json";
    } else {
      return "";
    }
  }

  String getEncoding() {
    return "charset=UTF-8";
  }
}

class HttpBody {
  HttpBody();

  late DataFormat format;
  late Encoding encoding;
  late Uint8List data;

  factory HttpBody.json(String json) {
    var body = HttpBody();
    body.format = DataFormat.json;
    body.encoding = utf8;
    body.data = Uint8List.fromList(utf8.encode(json));
    return body;
  }

  @override
  String toString() {
    if (format == DataFormat.json) {
      return encoding.decode(data);
    }
    return super.toString();
  }

  bool isTextFormat() {
    return format == DataFormat.json || format == DataFormat.xml || format == DataFormat.html;
  }

  Object convert() {
    if (isTextFormat()) {
      return toString();
    } else {
      return data;
    }
  }
}

class HttpOutput {
  HttpOutput({required this.isOk, required this.statusCode, this.message, this.encoding, this.data});

  bool isOk;
  int statusCode;
  DataFormat? format;
  String? message;
  Encoding? encoding;
  Uint8List? data;
}
