import 'dart:convert';

import 'package:modular_system_core/mscore.dart';

class WebClient<T extends JsonSerializer<T>> {}

class ApiClient<T extends JsonSerializer<T>> extends WebClient<T> {
  ApiClient({required this.uri, required this.controllerName, required this.constructor, this.token = ""}) {
    setSessionToken();
  }

  Uri uri;
  String controllerName;
  String token;
  JsonObjectCreator<T> constructor;
  ConcreteHttp client = ConcreteHttp();

  void setSessionToken() {
    uri = uri.addQueryParam("app-session", token);
  }

  // CREATE
  /// Creates an instance of T and returns its ID.
  Future<StringDTO> createAsync(T data) async {
    var req = HttpInput(
      uri: uri.replace(
        path: "api/$controllerName",
      ),
      method: HttpMethod.post,
      body: HttpBody.json(data.toJson()),
    );
    var resp = await client.sendAsync(req);
    var jsonMap = handleDataResponse(resp);
    return StringDTO.fromJson(jsonMap);
  }

  // READ
  Future<T> getByIdAsync(String id) async {
    var req = HttpInput(
      uri: uri.replace(
        path: "api/$controllerName",
        queryParameters: {"id": id},
      ),
      method: HttpMethod.get,
    );
    var resp = await client.sendAsync(req);
    var jsonMap = handleDataResponse(resp);
    return constructor(jsonMap);
  }

  Future<QueryList<T>> searchAsync(WebSearchRequest search) async {
    var req = HttpInput(
      uri: uri.replace(
        path: "api/$controllerName/query",
      ),
      method: HttpMethod.post,
      body: HttpBody.json(search.toJson()),
    );
    var resp = await client.sendAsync(req);
    var jsonMap = handleDataResponse(resp);
    return QueryList.fromJsonMap(jsonMap, constructor);
  }

  // UPDATE
  Future<BooleanDTO> updateAsync(WebUpdateRequest update) async {
    var req = HttpInput(
      uri: uri.replace(
        path: "api/$controllerName",
      ),
      method: HttpMethod.put,
      body: HttpBody.json(update.toJson()),
    );
    var resp = await client.sendAsync(req);
    var jsonMap = handleDataResponse(resp);
    return BooleanDTO.fromJson(jsonMap);
  }

  Future<BooleanDTO> overrideAsync(T data) async {
    var req = HttpInput(
      uri: uri.replace(
        path: "api/$controllerName/override",
      ),
      method: HttpMethod.put,
      body: HttpBody.json(data.toJson()),
    );
    var resp = await client.sendAsync(req);
    var jsonMap = handleDataResponse(resp);
    return BooleanDTO.fromJson(jsonMap);
  }

  // DELETE
  Future<BooleanDTO> deleteByIdAsync(String id) async {
    var req = HttpInput(
      uri: uri.replace(
        path: "api/$controllerName",
        queryParameters: {"id": id},
      ),
      method: HttpMethod.delete,
    );
    var resp = await client.sendAsync(req);
    var jsonMap = handleDataResponse(resp);
    return BooleanDTO.fromJson(jsonMap);
  }

  void handleNoDataResponse(HttpOutput resp) {
    if (!resp.isOk) {
      throw AppException.fromJson(getJsonFrom(resp), StackTrace.current);
    }
  }

  String getJsonFrom(HttpOutput resp) => utf8.decode(resp.data!.toList());

  Map<String, dynamic> getJsonMapFrom(HttpOutput resp) => jsonDecode(getJsonFrom(resp));

  Map<String, dynamic> handleDataResponse(HttpOutput resp) {
    if (resp.isOk && resp.data != null) {
      return getJsonMapFrom(resp);
    } else if (resp.data == null) {
      throw AppException("No data was returned from the request.", ExceptionCode.internal, StackTrace.current);
    } else {
      throw AppException.fromJson(getJsonFrom(resp), StackTrace.current);
    }
  }
}
