import 'package:flutter/material.dart';
import 'package:modular_system_core/mscore.dart';
import 'package:modular_system_core/src/model/disponsable.dart';

class DynamicFormState extends ReactiveObject<DynamicFormState> {
  bool isSubmitting = false;
  bool showErrors = false;
  List<String> errors = [];
}

// REFACTORING
class FormControllerOld implements Disposable {
  FormControllerOld({
    required this.viewModel,
    this.callback,
    this.asyncCallback,
  })  : assert(callback != null || asyncCallback != null),
        assert(!(callback != null && asyncCallback != null)) {
    subscribeToEvents();
  }

  AppViewModel viewModel;

  Function()? callback;
  Future Function()? asyncCallback;

  DynamicFormState state = DynamicFormState();
  Event validateEvent = Event();
  Event setEvent = Event();
  Event resetEvent = Event();

  Event submitEvent = Event();
  Event clearErrorsEvent = Event();
  SetEvent<bool> loaderEvent = SetEvent();
  SetEvent<String> submitErrorEvent = SetEvent();
  SetEvent<String> validationErrorEvent = SetEvent();
  SetEvent<List<String>> displayErrorsEvent = SetEvent();
  List<String> errors = [];

  @override
  void dispose() {
    unsubscribeToEvents();
  }

  void subscribeToEvents() {
    validationErrorEvent.subscribe(onValidationError);
  }

  void unsubscribeToEvents() {
    validationErrorEvent.unsubscribe(onValidationError);
  }

  void submitAsync() async {
    try {
      clearErrors();
      submitEvent.invoke();
      if (errors.isEmpty) {
        loaderEvent.invoke(true);
        if (callback != null) {
          callback!();
        }
        if (asyncCallback != null) {
          await asyncCallback!();
        }
      } else {
        showErrors();
      }
    } catch (e) {
      viewModel.errorHandler.handle(e);
    } finally {
      loaderEvent.invoke(false);
    }
  }

  void showErrors() {
    displayErrorsEvent.invoke(errors);
  }

  void clearErrors() {
    errors.clear();
    clearErrorsEvent.invoke();
  }

  void onValidationError(String error) {
    errors.add(error);
  }

  bool getIsFormValid() {
    clearErrors();
    submitEvent.invoke();
    if (errors.isEmpty) {
      return true;
    } else {
      return false;
    }
  }
}

class FieldState {
  String value = "";
  List<String> errors = [];
}

class FieldController {
  //new
  final FieldState state = FieldState();
  final Event resetEvent = Event();
  // old
  final GetEvent<String> getValueEvent = GetEvent();
  final GetEvent<bool> getIsValidEvent = GetEvent();

  void setValue(String value) => state.value = value;

  void addError(String error) {
    state.errors.add(error);
  }
}

class AppForm implements Disposable {
  AppForm({
    required this.controller,
  });

  final FormControllerOld controller;

  @override
  void dispose() {
    controller.dispose();
  }

  FormTextField textField({
    bool enabled = true,
    bool forbidSpaces = false,
    required String hintText,
    required String Function() getValue,
    required void Function(String data) setValue,
    FieldController? fieldController,
    RegExp? forbiddenCharacters,
    TextStyle? style,
    InputDecoration? decoration,
  }) {
    return FormTextField(
      enabled: enabled,
      forbidSpaces: forbidSpaces,
      hintText: hintText,
      formController: controller,
      fieldController: fieldController,
      getValue: getValue,
      setValue: setValue,
      forbiddenCharacters: forbiddenCharacters,
      style: style,
      decoration: decoration,
    );
  }

  FormIntField intField({
    bool enabled = true,
    int? minValue,
    int? maxValue,
    required String hintText,
    String? numberSymbol,
    String? trailingNumberSymbol,
    required int Function() getValue,
    required void Function(int data) setValue,
    FieldController? fieldController,
    TextStyle? style,
    InputDecoration? decoration,
  }) {
    return FormIntField(
      enabled: enabled,
      minValue: minValue,
      maxValue: maxValue,
      hintText: hintText,
      numberSymbol: numberSymbol,
      trailingNumberSymbol: trailingNumberSymbol,
      formController: controller,
      fieldController: fieldController,
      getValue: getValue,
      setValue: setValue,
      style: style,
      decoration: decoration,
    );
  }

  FormDoubleField doubleField({
    bool enabled = true,
    required int decimals,
    double? minValue,
    double? maxValue,
    required String hintText,
    String? numberSymbol,
    String? trailingNumberSymbol,
    required double Function() getValue,
    required void Function(double data) setValue,
    FieldController? fieldController,
    TextStyle? style,
    InputDecoration? decoration,
  }) {
    return FormDoubleField(
      enabled: enabled,
      decimals: decimals,
      minValue: minValue,
      maxValue: maxValue,
      hintText: hintText,
      numberSymbol: numberSymbol,
      trailingNumberSymbol: trailingNumberSymbol,
      formController: controller,
      fieldController: fieldController,
      getValue: getValue,
      setValue: setValue,
      style: style,
      decoration: decoration,
    );
  }

  FormDropdownField dropdownField({
    bool enabled = true,
    bool forbidSpaces = false,
    required String hintText,
    required int Function() getValue,
    required void Function(int data) setValue,
    FieldController? fieldController,
    required Map<int, String> dropdownItems,
    RegExp? forbiddenCharacters,
    TextStyle? style,
    InputDecoration? decoration,
  }) {
    return FormDropdownField(
      enabled: enabled,
      hintText: hintText,
      formController: controller,
      fieldController: fieldController,
      getValue: getValue,
      setValue: setValue,
      dropdownItems: dropdownItems,
      style: style,
      decoration: decoration,
    );
  }

  FormCheckboxField checkboxField({
    bool enabled = true,
    required String hintText,
    required bool Function() getValue,
    required void Function(bool data) setValue,
    FieldController? fieldController,
  }) {
    return FormCheckboxField(
      enabled: enabled,
      hintText: hintText,
      formController: controller,
      fieldController: fieldController,
      getValue: getValue,
      setValue: setValue,
    );
  }

  FormErrorDisplay errorDisplay({
    EdgeInsets? margin,
    EdgeInsets? padding = const EdgeInsets.all(10),
  }) {
    return FormErrorDisplay(
      formController: controller,
      margin: margin,
      padding: padding,
    );
  }
}

// Error display

class FormErrorDisplay extends StatefulWidget {
  // ignore: use_key_in_widget_constructors
  const FormErrorDisplay({
    required this.formController,
    this.margin,
    this.padding = const EdgeInsets.all(10),
  });

  final FormControllerOld formController;
  final EdgeInsets? margin;
  final EdgeInsets? padding;

  @override
  State<FormErrorDisplay> createState() {
    // ignore: no_logic_in_create_state
    return _FormErrorDisplayState(
      formController: formController,
      margin: margin,
      padding: padding,
    );
  }
}

class _FormErrorDisplayState extends State<FormErrorDisplay> {
  _FormErrorDisplayState({
    required this.formController,
    this.margin,
    this.padding,
  }) {
    formController.clearErrorsEvent.subscribe(onClearErrors);
    formController.displayErrorsEvent.subscribe(onDisplayErrors);
  }

  @override
  void dispose() {
    formController.clearErrorsEvent.unsubscribe(onClearErrors);
    formController.displayErrorsEvent.unsubscribe(onDisplayErrors);
    super.dispose();
  }

  final FormControllerOld formController;
  final EdgeInsets? margin;
  final EdgeInsets? padding;
  final List<String> errors = [];

  @override
  Widget build(BuildContext context) {
    if (errors.isEmpty) {
      return Container();
    }
    return Container(
      margin: margin,
      padding: padding,
      child: Column(
        children: [
          Row(
            children: [
              Icon(
                Icons.warning_amber,
                color: formController.viewModel.config.colorScheme.onError,
                size: formController.viewModel.config.fontSize.data,
              ),
              const VerticalDivider(),
              Flexible(
                child: Text(
                  "Corrija os seguintes erros antes de tentar novamente:",
                  maxLines: 2,
                  style: formController.viewModel.styleProvider.auto(FontType.onError),
                ),
              ),
            ],
          ),
          Text(
            getErrorText(),
            maxLines: 50,
            style: formController.viewModel.styleProvider.auto(FontType.onError),
          ),
        ],
      ),
    );
  }

  String getErrorText() {
    var str = "";
    for (var error in errors) {
      str = '$str\n- $error';
    }
    return str;
  }

  void onDisplayErrors(List<String> errors) {
    setState(() {
      this.errors.addAll(errors);
    });
  }

  void onClearErrors() {
    setState(() {
      errors.clear();
    });
  }
}

// FIELDS

mixin FormFieldOld<T extends Object> {
  abstract T Function() getValue;
  abstract void Function(T data) setValue;
}

// String field
class FormTextField extends StatefulWidget {
  // ignore: use_key_in_widget_constructors
  const FormTextField({
    this.enabled = true,
    this.forbidSpaces = false,
    required this.hintText,
    this.initialValue,
    required this.formController,
    this.fieldController,
    required this.getValue,
    required this.setValue,
    this.forbiddenCharacters,
    this.style,
    this.decoration,
  });

  final bool enabled;
  final bool forbidSpaces;
  final String hintText;
  final String? initialValue;
  final FormControllerOld formController;
  final FieldController? fieldController;
  final void Function(String data) setValue;
  final String Function() getValue;
  final RegExp? forbiddenCharacters;
  final TextStyle? style;
  final InputDecoration? decoration;

  @override
  State<FormTextField> createState() {
    // ignore: no_logic_in_create_state
    return _FormTextFieldState(
      enabled: enabled,
      forbidSpaces: forbidSpaces,
      hintText: hintText,
      initialValue: initialValue,
      formController: formController,
      fieldController: fieldController,
      getValue: getValue,
      setValue: setValue,
      forbiddenCharacters: forbiddenCharacters ?? RegExp('a^'),
      style: style,
      decoration: decoration,
    );
  }
}

class _FormTextFieldState extends State<FormTextField> {
  _FormTextFieldState({
    required this.enabled,
    required this.forbidSpaces,
    required this.hintText,
    this.initialValue,
    required this.formController,
    this.fieldController,
    required this.getValue,
    required this.setValue,
    this.forbiddenCharacters,
    this.style,
    this.decoration,
  }) {
    subscribeToEvents();
  }

  // injected
  bool enabled;
  bool forbidSpaces;
  String hintText;
  String? initialValue;
  FormControllerOld formController;
  FieldController? fieldController;
  void Function(String data) setValue;
  String Function() getValue;
  RegExp? forbiddenCharacters;
  TextStyle? style;
  InputDecoration? decoration;
  // state
  String errorMessage = "";
  late TextEditingController controller = TextEditingController(
    text: initialValue ?? getValue.call(),
  );

  @override
  void dispose() {
    unsubscribeToEvents();
    super.dispose();
  }

  void subscribeToEvents() {
    formController.submitEvent.subscribe(onSubmit);
    formController.resetEvent.subscribe(onReset);
    if (fieldController != null) {
      fieldController!.resetEvent.subscribe(onReset);
      fieldController!.getValueEvent.subscribe(onGetValue);
      fieldController!.getIsValidEvent.subscribe(onGetIsValid);
    }
  }

  void unsubscribeToEvents() {
    formController.submitEvent.unsubscribe(onSubmit);
    formController.resetEvent.unsubscribe(onReset);
    if (fieldController != null) {
      fieldController!.resetEvent.unsubscribe(onReset);
      fieldController!.getValueEvent.unsubscribe(onGetValue);
      fieldController!.getIsValidEvent.unsubscribe(onGetIsValid);
    }
  }

  @override
  Widget build(BuildContext context) {
    return Container(
      margin: const EdgeInsets.symmetric(
        vertical: 20,
      ),
      child: TextField(
        controller: controller,
        enabled: enabled,
        style: style ?? defaultTextStyle(),
        decoration: decoration ?? defaultInputDecoration(),
        onChanged: onChanged,
      ),
    );
  }

  TextStyle defaultTextStyle() {
    return formController.viewModel.styleProvider.auto(FontType.onBackground);
  }

  InputDecoration defaultInputDecoration() {
    return InputDecoration(
      labelText: hintText,
      floatingLabelBehavior: FloatingLabelBehavior.always,
      labelStyle: errorMessage.isNotEmpty
          ? formController.viewModel.styleProvider.auto(FontType.onError, FontSize.label)
          : formController.viewModel.styleProvider.auto(FontType.onBackground, FontSize.label),
      hintText: !enabled ? "Desabilitado" : null,
      //hintStyle: TextStyle(fontStyle: FontStyle.italic),
      helperText: errorMessage,
      helperStyle: formController.viewModel.styleProvider.auto(FontType.onError),
      focusedBorder: OutlineInputBorder(
        borderSide: BorderSide(
          color: errorMessage.isNotEmpty
              ? formController.viewModel.config.colorScheme.onError
              : formController.viewModel.config.colorScheme.onBackground,
        ),
      ),
      enabledBorder: OutlineInputBorder(
        borderSide: BorderSide(
          color: errorMessage.isNotEmpty
              ? formController.viewModel.config.colorScheme.onError
              : formController.viewModel.config.colorScheme.onBackground,
        ),
      ),
      disabledBorder: const OutlineInputBorder(
        borderSide: BorderSide(
          color: Colors.orange,
        ),
      ),
    );
  }

  String getFieldValue() {
    return controller.text.trim();
  }

  void showError(String message) {
    setState(() {
      errorMessage = message;
    });
  }

  void hideError() {
    setState(() {
      errorMessage = "";
    });
  }

  bool getIsInputValid() {
    if (forbiddenCharacters!.hasMatch(getFieldValue()) || getFieldValue().contains(" ") && forbidSpaces) {
      return false;
    }
    return true;
  }

  List<String> getForbbidenCharacters() {
    var list = <String>[];
    for (var match in forbiddenCharacters!.allMatches(getFieldValue())) {
      var char = match[0].toString();
      list.add(char.isNotEmpty ? char : "*espaço em branco");
    }
    if (RegExp(" ").allMatches(getFieldValue()).isNotEmpty) {
      list.add("espaços em branco");
    }
    return list;
  }

  String getErrorMessage(bool withName) {
    var message = "";
    if (forbiddenCharacters!.hasMatch(getFieldValue()) || getFieldValue().contains(" ") && forbidSpaces) {
      message = "Caracteres proibidos foram encontrados, remova os seguintes caracteres: [";
      var forbbidenCharacters = getForbbidenCharacters();
      var length = forbbidenCharacters.length;
      for (var i = 0; i < length; i++) {
        var char = forbbidenCharacters[i];

        if (i == 0 && length == 1) {
          message = "$message $char ].";
        } else if (i == 0) {
          message = "$message $char";
        } else if (i + 1 == length) {
          message = "$message, $char ].";
        } else {
          message = "$message, $char";
        }
      }
    } else {
      message = "Erro desconhecido.";
    }
    if (withName) {
      message = "$hintText: $message";
    }
    return message;
  }

  void onChanged(String value) {
    if (getIsInputValid()) {
      hideError();
      setValue(getFieldValue());
    } else {
      showError(getErrorMessage(false));
    }
  }

  void onSubmit() {
    if (!getIsInputValid()) {
      showError(getErrorMessage(false));
      formController.validationErrorEvent.invoke(getErrorMessage(true));
    } else {
      hideError();
      setValue(getFieldValue());
    }
  }

  void onReset() {
    controller.clear();
  }

  String onGetValue() {
    return getFieldValue();
  }

  bool onGetIsValid() {
    return getIsInputValid();
  }
}

// Int field
class FormIntField extends StatefulWidget {
  // ignore: use_key_in_widget_constructors
  const FormIntField({
    this.enabled = true,
    this.minValue,
    this.maxValue,
    required this.hintText,
    this.numberSymbol,
    this.trailingNumberSymbol,
    required this.formController,
    this.fieldController,
    required this.getValue,
    required this.setValue,
    this.style,
    this.decoration,
  });

  final bool enabled;
  final int? minValue;
  final int? maxValue;
  final String hintText;
  final String? numberSymbol;
  final String? trailingNumberSymbol;
  final FormControllerOld formController;
  final FieldController? fieldController;
  final void Function(int data) setValue;
  final int Function() getValue;
  final TextStyle? style;
  final InputDecoration? decoration;

  @override
  State<FormIntField> createState() {
    // ignore: no_logic_in_create_state
    return _FormIntFieldState(
      enabled: enabled,
      minValue: minValue,
      maxValue: maxValue,
      hintText: hintText,
      numberSymbol: numberSymbol,
      trailingNumberSymbol: trailingNumberSymbol,
      formController: formController,
      fieldController: fieldController,
      getValue: getValue,
      setValue: setValue,
      style: style,
      decoration: decoration,
    );
  }
}

class _FormIntFieldState extends State<FormIntField> {
  _FormIntFieldState({
    required this.enabled,
    this.minValue,
    this.maxValue,
    required this.hintText,
    this.numberSymbol,
    this.trailingNumberSymbol,
    required this.formController,
    this.fieldController,
    required this.getValue,
    required this.setValue,
    this.style,
    this.decoration,
  }) {
    subscribeToEvents();
    init();
  }

  // injected
  bool enabled;
  int? minValue;
  int? maxValue;
  String hintText;
  String? numberSymbol;
  String? trailingNumberSymbol;
  FormControllerOld formController;
  FieldController? fieldController;
  void Function(int data) setValue;
  int Function() getValue;
  TextStyle? style;
  InputDecoration? decoration;
  // state
  String errorMessage = "";
  TextEditingController controller = TextEditingController();

  @override
  void dispose() {
    unsubscribeToEvents();
    super.dispose();
  }

  void init() {
    var initialValue = getValue.call().toString();
    controller.text = initialValue;
  }

  void subscribeToEvents() {
    formController.submitEvent.subscribe(onSubmit);
    formController.resetEvent.subscribe(onReset);
  }

  void unsubscribeToEvents() {
    formController.submitEvent.unsubscribe(onSubmit);
    formController.resetEvent.unsubscribe(onReset);
  }

  @override
  Widget build(BuildContext context) {
    return Container(
      margin: const EdgeInsets.symmetric(
        vertical: 20,
      ),
      child: TextField(
        controller: controller,
        enabled: enabled,
        style: style ?? defaultTextStyle(),
        decoration: decoration ?? defaultInputDecoration(),
        onChanged: onChanged,
      ),
    );
  }

  TextStyle defaultTextStyle() {
    return formController.viewModel.styleProvider.auto(FontType.onBackground);
  }

  InputDecoration defaultInputDecoration() {
    return InputDecoration(
      labelText: hintText,
      floatingLabelBehavior: FloatingLabelBehavior.always,
      labelStyle: errorMessage.isNotEmpty
          ? formController.viewModel.styleProvider.auto(FontType.onError, FontSize.label)
          : formController.viewModel.styleProvider.auto(FontType.onBackground, FontSize.label),
      hintText: !enabled ? "Desabilitado" : null,
      //hintStyle: TextStyle(fontStyle: FontStyle.italic),
      helperText: errorMessage,
      helperStyle: formController.viewModel.styleProvider.auto(FontType.onError),
      focusedBorder: OutlineInputBorder(
        borderSide: BorderSide(
          color: errorMessage.isNotEmpty
              ? formController.viewModel.config.colorScheme.onError
              : formController.viewModel.config.colorScheme.onBackground,
        ),
      ),
      enabledBorder: OutlineInputBorder(
        borderSide: BorderSide(
          color: errorMessage.isNotEmpty
              ? formController.viewModel.config.colorScheme.onError
              : formController.viewModel.config.colorScheme.onBackground,
        ),
      ),
      disabledBorder: const OutlineInputBorder(
        borderSide: BorderSide(
          color: Colors.orange,
        ),
      ),
    );
  }

  String getRawValue() {
    return controller.text.trim();
  }

  int getFieldValue() {
    var value = controller.text.trim().replaceAll(RegExp("[^0-9]"), "");
    if (value.isNotEmpty) {
      return int.parse(value);
    }
    return 0;
  }

  void showError(String message) {
    setState(() {
      errorMessage = message;
    });
  }

  void hideError() {
    setState(() {
      errorMessage = "";
    });
  }

  bool getIsInputValid() {
    if (minValue != null) {
      if (getFieldValue() < minValue!) {
        return false;
      }
    }
    if (maxValue != null) {
      if (getFieldValue() > maxValue!) {
        return false;
      }
    }
    return true;
  }

  String getErrorMessage(bool withName) {
    var message = "";
    if (!getIsInputValid()) {
      message = "Valor inválido";
    } else {
      message = "Erro desconhecido.";
    }
    if (minValue != null) {
      if (getFieldValue() <= minValue!) {
        message = "O valor deve ser maior ou igual a $minValue.";
      }
    }
    if (maxValue != null) {
      if (getFieldValue() > maxValue!) {
        message = "O valor deve ser menor ou igual a $minValue.";
      }
    }
    if (withName) {
      message = "$hintText: $message";
    }
    return message;
  }

  void onChanged(String value) {
    try {
      var str = '${getFieldValue()}';
      if (numberSymbol != null) {
        str = "$numberSymbol $str";
      }
      if (trailingNumberSymbol != null) {
        str = "$str $trailingNumberSymbol";
      }
      controller.text = str;
      controller.selection = TextSelection.fromPosition(TextPosition(offset: controller.text.length));
      if (!getIsInputValid()) {
        showError(getErrorMessage(false));
      } else {
        hideError();
        setValue(getFieldValue());
      }
    } catch (e) {
      formController.viewModel.errorHandler.handle(e);
    }
  }

  void onSubmit() {
    if (!getIsInputValid()) {
      formController.validationErrorEvent.invoke(getErrorMessage(true));
    } else {
      setValue(getFieldValue());
    }
  }

  void onReset() {
    controller.text = "";
    onChanged("");
  }
}

// Double field
class FormDoubleField extends StatefulWidget {
  // ignore: use_key_in_widget_constructors
  const FormDoubleField({
    this.enabled = true,
    required this.decimals,
    this.minValue,
    this.maxValue,
    required this.hintText,
    this.numberSymbol,
    this.trailingNumberSymbol,
    required this.formController,
    this.fieldController,
    required this.getValue,
    required this.setValue,
    this.style,
    this.decoration,
  });

  final bool enabled;
  final int decimals;
  final double? minValue;
  final double? maxValue;
  final String hintText;
  final String? numberSymbol;
  final String? trailingNumberSymbol;
  final FormControllerOld formController;
  final FieldController? fieldController;
  final void Function(double data) setValue;
  final double Function() getValue;
  final TextStyle? style;
  final InputDecoration? decoration;

  @override
  State<FormDoubleField> createState() {
    // ignore: no_logic_in_create_state
    return _FormDoubleFieldState();
  }
}

class _FormDoubleFieldState extends State<FormDoubleField> {
  _FormDoubleFieldState();

  FieldController fieldController = FieldController();

  // state
  String errorMessage = "";
  late CurrencyText currencyText = CurrencyText(
    symbol: widget.numberSymbol,
    trailingSymbol: widget.trailingNumberSymbol,
    decimals: widget.decimals,
  );
  late TextEditingController controller = TextEditingController(
    text: currencyText.toString(),
  );

  @override
  void initState() {
    super.initState();
    subscribeToEvents();
    controller.text = widget.getValue().toStringAsFixed(widget.decimals);
  }

  @override
  void dispose() {
    unsubscribeToEvents();
    super.dispose();
  }

  void subscribeToEvents() {
    widget.formController.submitEvent.subscribe(onSubmit);
    widget.formController.resetEvent.subscribe(onReset);
  }

  void unsubscribeToEvents() {
    widget.formController.submitEvent.unsubscribe(onSubmit);
    widget.formController.resetEvent.unsubscribe(onReset);
  }

  @override
  Widget build(BuildContext context) {
    return Container(
      margin: const EdgeInsets.symmetric(
        vertical: 20,
      ),
      child: TextField(
        controller: controller,
        enabled: widget.enabled,
        style: widget.style ?? defaultTextStyle(),
        decoration: widget.decoration ?? defaultInputDecoration(),
        onChanged: onChanged,
      ),
    );
  }

  TextStyle defaultTextStyle() {
    return widget.formController.viewModel.styleProvider.auto(FontType.onBackground);
  }

  InputDecoration defaultInputDecoration() {
    return InputDecoration(
      labelText: widget.hintText,
      floatingLabelBehavior: FloatingLabelBehavior.always,
      labelStyle: errorMessage.isNotEmpty
          ? widget.formController.viewModel.styleProvider.auto(FontType.onError, FontSize.label)
          : widget.formController.viewModel.styleProvider.auto(FontType.onBackground, FontSize.label),
      hintText: !widget.enabled ? "Desabilitado" : null,
      //hintStyle: TextStyle(fontStyle: FontStyle.italic),
      helperText: errorMessage,
      helperStyle: widget.formController.viewModel.styleProvider.auto(FontType.onError),
      focusedBorder: OutlineInputBorder(
        borderSide: BorderSide(
          color: errorMessage.isNotEmpty
              ? widget.formController.viewModel.config.colorScheme.onError
              : widget.formController.viewModel.config.colorScheme.onBackground,
        ),
      ),
      enabledBorder: OutlineInputBorder(
        borderSide: BorderSide(
          color: errorMessage.isNotEmpty
              ? widget.formController.viewModel.config.colorScheme.onError
              : widget.formController.viewModel.config.colorScheme.onBackground,
        ),
      ),
      disabledBorder: const OutlineInputBorder(
        borderSide: BorderSide(
          color: Colors.orange,
        ),
      ),
    );
  }

  double getFieldValue() {
    var value = controller.text.trim().replaceAll(RegExp("[^0-9]"), "");
    if (value.isNotEmpty) {
      return double.parse(value);
    }
    return 0;
  }

  bool getIsInputValid() {
    if (widget.minValue != null) {
      if (getFieldValue() < widget.minValue!) {
        return false;
      }
    }
    if (widget.maxValue != null) {
      if (getFieldValue() > widget.maxValue!) {
        return false;
      }
    }
    return true;
  }

  String getErrorMessage(bool withName) {
    var message = "";
    if (!getIsInputValid()) {
      message = "Valor inválido";
    } else {
      message = "Erro desconhecido.";
    }
    if (widget.minValue != null) {
      if (getFieldValue() <= widget.minValue!) {
        message = "O valor deve ser maior ou igual a ${widget.minValue}.";
      }
    }
    if (widget.maxValue != null) {
      if (getFieldValue() > widget.maxValue!) {
        message = "O valor deve ser menor ou igual a ${widget.minValue}.";
      }
    }
    if (withName) {
      message = "${widget.hintText}: $message";
    }
    return message;
  }

  void onChanged(String data) {
    var numbers = data.replaceAll(RegExp("[^0-9]"), ""); // '0009'
    var chars = numbers.characters.toList(); // [0, 0, 0, 9]
    var minCharCount = widget.decimals + 1;
    chars = trimZeroes(chars); // [9]
    currencyText.clear();
    while (chars.length < minCharCount) {
      chars.insert(0, '0');
    }
    // chars = [0, 0, 9]
    for (var index = 0; index < widget.decimals; index++) {
      currencyText.addNumber(
        Number(
          decimalPosition: -(widget.decimals - index),
          value: int.parse(chars.reversed.elementAt(index)),
        ),
      );
    }
    // chars = [1, 5, 0,| 0 ,0]
    for (var index = 0; index < chars.length - widget.decimals; index++) {
      currencyText.addNumber(
        Number(
          decimalPosition: chars.length - widget.decimals - index,
          value: int.parse(chars.elementAt(index)),
        ),
      );
    }
    controller.text = currencyText.toString();
    if (widget.trailingNumberSymbol != null) {
      controller.selection = TextSelection.fromPosition(TextPosition(offset: controller.text.length - (widget.numberSymbol!.length + 1)));
    } else {
      controller.selection = TextSelection.fromPosition(TextPosition(offset: controller.text.length));
    }
    widget.setValue(currencyText.toDouble());
  }

  List<String> trimZeroes(List<String> chars) {
    while (chars.isNotEmpty && chars[0] == '0') {
      chars.removeAt(0);
    }
    return chars;
  }

  void onSubmit() {
    if (!getIsInputValid()) {
      widget.formController.validationErrorEvent.invoke(getErrorMessage(true));
    } else {
      widget.setValue(currencyText.toDouble());
    }
  }

  void onReset() {
    controller.text = "";
    onChanged("");
  }
}

// Dropdown dield
class FormDropdownField extends StatefulWidget {
  // ignore: use_key_in_widget_constructors
  const FormDropdownField({
    this.enabled = true,
    required this.hintText,
    required this.formController,
    this.fieldController,
    required this.getValue,
    required this.setValue,
    required this.dropdownItems,
    this.style,
    this.decoration,
  });

  final bool enabled;
  final String hintText;
  final FormControllerOld formController;
  final FieldController? fieldController;
  final void Function(int data) setValue;
  final int Function() getValue;
  final Map<int, String> dropdownItems;
  final TextStyle? style;
  final InputDecoration? decoration;

  @override
  State<FormDropdownField> createState() {
    // ignore: no_logic_in_create_state
    return _FormDropdownFieldState(
      enabled: enabled,
      hintText: hintText,
      formController: formController,
      fieldController: fieldController,
      getValue: getValue,
      setValue: setValue,
      dropdownItems: dropdownItems,
      style: style,
      decoration: decoration,
    );
  }
}

class _FormDropdownFieldState extends State<FormDropdownField> {
  _FormDropdownFieldState({
    required this.enabled,
    required this.hintText,
    required this.formController,
    this.fieldController,
    required this.getValue,
    required this.setValue,
    required this.dropdownItems,
    this.style,
    this.decoration,
  }) {
    dropdownValue = getValue();
  }

  // injected
  bool enabled;
  String hintText;
  FormControllerOld formController;
  FieldController? fieldController;
  final void Function(int data) setValue;
  final int Function() getValue;
  Map<int, String> dropdownItems;
  TextStyle? style;
  InputDecoration? decoration;
  // state
  int dropdownValue = 0;
  String errorMessage = "";

  @override
  Widget build(BuildContext context) {
    return Container(
      margin: const EdgeInsets.symmetric(
        vertical: 20,
      ),
      child: Column(
        crossAxisAlignment: CrossAxisAlignment.start,
        children: [
          Text(
            hintText,
            style: formController.viewModel.styleProvider.auto(FontType.onBackground, FontSize.label),
          ),
          DropdownButtonHideUnderline(
            child: DropdownButton(
              isExpanded: true,
              value: dropdownValue,
              focusNode: FocusNode(canRequestFocus: false),
              items: getDropdownItemsList(),
              onChanged: (value) {
                setState(() {
                  dropdownValue = value as int;
                  setValue.call(dropdownValue);
                });
              },
              style: formController.viewModel.styleProvider.auto(FontType.onBackground),
              hint: Text(
                hintText,
                style: formController.viewModel.styleProvider.auto(FontType.onBackground, FontSize.label),
              ),
            ),
          ),
        ],
      ),
    );
  }

  List<DropdownMenuItem<int>> getDropdownItemsList() {
    var list = <DropdownMenuItem<int>>[];
    for (var i = 0; i < dropdownItems.length; i++) {
      var item = dropdownItems[i];
      list.add(
        DropdownMenuItem(
          child: Text(item!),
          value: i,
        ),
      );
    }
    return list;
  }

  void onReset() {
    dropdownValue = 0;
  }
}

// Checkbox field
class FormCheckboxField extends StatefulWidget {
  // ignore: use_key_in_widget_constructors
  const FormCheckboxField({
    this.enabled = true,
    required this.hintText,
    required this.formController,
    this.fieldController,
    required this.getValue,
    required this.setValue,
  });

  final bool enabled;
  final String hintText;
  final FormControllerOld formController;
  final FieldController? fieldController;
  final void Function(bool data) setValue;
  final bool Function() getValue;

  @override
  State<FormCheckboxField> createState() {
    // ignore: no_logic_in_create_state
    return _FormCheckboxFieldState(
      enabled: enabled,
      hintText: hintText,
      formController: formController,
      fieldController: fieldController,
      getValue: getValue,
      setValue: setValue,
    );
  }
}

class _FormCheckboxFieldState extends State<FormCheckboxField> {
  _FormCheckboxFieldState({
    required this.enabled,
    required this.hintText,
    required this.formController,
    this.fieldController,
    required this.getValue,
    required this.setValue,
  });

  bool enabled;
  String hintText;
  FormControllerOld formController;
  FieldController? fieldController;
  void Function(bool data) setValue;
  bool Function() getValue;
  // inner state
  late bool checkboxValue = getValue.call();

  @override
  Widget build(BuildContext context) {
    return Container(
      margin: const EdgeInsets.symmetric(
        vertical: 20,
      ),
      child: Column(
        crossAxisAlignment: CrossAxisAlignment.start,
        children: [
          Text(
            hintText,
            style: formController.viewModel.styleProvider.auto(FontType.onBackground, FontSize.label),
          ),
          Checkbox(
            value: checkboxValue,
            activeColor: formController.viewModel.config.colorScheme.tertiary,
            splashRadius: formController.viewModel.config.fontSize.data,
            onChanged: (value) {
              setState(() {
                checkboxValue = value!;
                setValue(value);
              });
            },
          ),
        ],
      ),
    );
  }
}

// other stuff
class Number {
  Number({
    required this.decimalPosition,
    required this.value,
  });

  int decimalPosition;
  int value;

  @override
  String toString() {
    return value.toString();
  }

  double toDouble() {
    var multiplier = "";
    // 1 -2 = 0,01 = x * 0.01
    if (decimalPosition.isNegative) {
      multiplier = "0.";
      for (var i = 0; i < decimalPosition * -1 - 1; i++) {
        multiplier = "${multiplier}0";
      }
      multiplier = "${multiplier}1";
      return value * double.parse(multiplier);
    } else {
      multiplier = "1";
      for (var i = 0; i < decimalPosition - 1; i++) {
        multiplier = "${multiplier}0";
      }
      return value * double.parse(multiplier);
    }
  }
}

class DoubleText {
  DoubleText(this.decimals);

  int decimals;
  List<Number> numbers = [];

  DoubleText clear() {
    numbers.clear();
    return this;
  }

  DoubleText addNumber(Number number) {
    var isDuplicatedPosition = numbers.where((x) => x.decimalPosition == number.decimalPosition).isNotEmpty;
    if (isDuplicatedPosition) {
      throw AppException("Trying to add two numbers to the same decimal position");
    } else if (number.decimalPosition == 0) {
      throw AppException("Trying to add number with 0 as decimal position");
    }
    numbers.add(number);
    return this;
  }

  @override
  String toString() {
    var str = '';
    var integerPart = '';
    var fractionPart = '';
    var integers = numbers.where((element) => element.decimalPosition > 0).toList();
    var fractions = numbers.where((element) => element.decimalPosition < 0).toList();
    if (integers.isEmpty) {
      integers.add(Number(decimalPosition: 1, value: 0));
    }
    for (var i = 1; i <= decimals; i++) {
      if (fractions.where((x) => x.decimalPosition == -i).isEmpty) {
        fractions.add(Number(decimalPosition: -i, value: 0));
      }
    }
    integers.sort(sort);
    fractions.sort(reveresedSort);
    for (var number in integers) {
      integerPart = '${number.toString()}$integerPart';
    }
    for (var number in fractions) {
      fractionPart = '$fractionPart${number.toString()}';
    }
    var _integerPart = integerPart;
    for (var i = 0; i < integerPart.length; i++) {
      if (i % 3 == 0 && i != 0) {
        var chars = _integerPart.characters.toList();
        chars.insert(integerPart.length - i, '.');
        _integerPart = chars.toString().replaceAll(RegExp("[^0-9.]"), "");
      }
    }
    integerPart = _integerPart;
    return '$integerPart,$fractionPart';
  }

  int sort(Number a, Number b) {
    return a.decimalPosition > b.decimalPosition ? 1 : 0;
  }

  int reveresedSort(Number a, Number b) {
    return a.decimalPosition < b.decimalPosition ? 1 : 0;
  }

  double toDouble() {
    var value = 0.0;
    for (var number in numbers) {
      value = value + number.toDouble();
    }
    return value;
  }
}

class CurrencyText extends DoubleText {
  CurrencyText({
    required int decimals,
    this.symbol,
    this.trailingSymbol,
  }) : super(decimals);

  String? symbol;
  String? trailingSymbol;

  @override
  String toString() {
    var str = super.toString();
    if (symbol != null) {
      str = '$symbol $str';
    }
    if (trailingSymbol != null) {
      str = '$str $trailingSymbol';
    }
    return str;
  }
}

//

class FieldDescriptor<Entity, Field> {
  FieldDescriptor({
    required this.name,
    required this.getValue,
    required this.setValue,
  });

  String name;
  Field Function(Entity instance) getValue;
  Entity Function(Field value) setValue;
}

class FormFieldProto extends StatefulWidget {
  const FormFieldProto({
    Key? key,
    required this.controller,
    required this.descriptor,
  }) : super(key: key);

  final FormControllerOld controller;
  final FieldDescriptor descriptor;

  @override
  State<FormFieldProto> createState() {
    // ignore: no_logic_in_create_state
    return FormFieldProtoState(
      controller: controller,
      descriptor: descriptor,
    );
  }
}

class FormFieldProtoState<FieldType extends FormFieldProto> extends State<FieldType> {
  FormFieldProtoState({
    required this.controller,
    required this.descriptor,
  });

  final FormControllerOld controller;
  final FieldDescriptor descriptor;

  @override
  Widget build(BuildContext context) {
    return Container();
  }
}
