import 'package:flutter/material.dart';
import 'package:modular_system_core/mscore.dart';
import 'package:modular_system_core/src/model/disponsable.dart';

class FormController implements Disposable {
  FormController({
    required this.viewModel,
    this.callback,
    this.asyncCallback,
  })  : assert(callback != null || asyncCallback != null),
        assert(!(callback != null && asyncCallback != null));

  AppViewModel viewModel;
  Function()? callback;
  Future Function()? asyncCallback;
  DynamicFormState state = DynamicFormState();
  Event validateEvent = Event();
  Event setEvent = Event();
  Event resetEvent = Event();

  @override
  void dispose() {
    // TODO: implement dispose
  }

  void submitAsync() async {
    try {
      if (formIsValid()) {
        if (callback != null) {
          callback!();
        }
        if (asyncCallback != null) {
          await asyncCallback!();
        }
      }
    } catch (e) {
      viewModel.errorHandler.handle(e);
    }
  }

  void addError(String error) {
    if (!state.errors.contains(error)) {
      state.errors.add(error);
    }
  }

  bool formIsValid() {
    state.errors.clear();
    validateEvent.invoke();
    return state.errors.isEmpty;
  }
}

class DynamicForm {
  DynamicForm({required this.viewModel});

  final AppViewModel viewModel;
  late String value = "";
  late FormController controller = FormController(viewModel: viewModel, callback: () {});

  Widget test() {
    return DynamicFormField(
      formController: controller,
      getter: () => value,
      setter: (value) => value = value,
      onChanged: (controller) {},
    );
  }
}

class DynamicFormField extends StatefulWidget {
  const DynamicFormField({
    Key? key,
    required this.formController,
    required this.getter,
    required this.setter,
    this.enabled = true,
    this.forbidSpaces = false,
    this.validateOnChanged = true,
    this.margin,
    this.style,
    this.decoration,
    this.forbiddenCharacters,
    this.onChanged,
    this.onValidate,
    this.fieldController,
  }) : super(key: key);

  final FormController formController;
  final String Function() getter;
  final void Function(String data) setter;
  // optional
  final bool enabled;
  final bool forbidSpaces;
  final bool validateOnChanged;
  final EdgeInsets? margin;
  final TextStyle? style;
  final InputDecoration? decoration;
  final RegExp? forbiddenCharacters;
  final void Function(FieldController)? onChanged;
  final void Function(FieldController)? onValidate;
  final FieldController? fieldController;

  @override
  State<DynamicFormField> createState() {
    return _DynamicFormFieldState();
  }
}

class _DynamicFormFieldState extends State<DynamicFormField> {
  late String errorMessage = "";
  late TextEditingController controller = TextEditingController();
  late FieldController fieldController;

  @override
  void initState() {
    super.initState();
    controller.text = widget.getter.call();
    fieldController = widget.fieldController ?? FieldController();
    subscribeToEvents();
  }

  @override
  void dispose() {
    unsubscribeToEvents();
    super.dispose();
  }

  void subscribeToEvents() {
    widget.formController.setEvent.subscribe(onSet);
    widget.formController.resetEvent.subscribe(onReset);
    widget.formController.validateEvent.subscribe(onValidate);
  }

  void unsubscribeToEvents() {
    widget.formController.setEvent.unsubscribe(onSet);
    widget.formController.resetEvent.unsubscribe(onReset);
    widget.formController.validateEvent.unsubscribe(onValidate);
  }

  void validate() {
    if (containsForbiddenCharacters()) {
      fieldController.addError(getForbiddenCharactersError());
    }
    for (var error in fieldController.state.errors) {
      widget.formController.addError(error);
    }
    widget.formController.state.notify();
  }

  void setValue() => widget.setter.call(getValue());

  bool containsForbiddenCharacters() {
    if (getValue().contains(" ") && widget.forbidSpaces) {
      return true;
    }
    if (widget.forbiddenCharacters != null && widget.forbiddenCharacters!.hasMatch(getValue())) {
      return true;
    }
    return false;
  }

  String getValue() {
    return fieldController.state.value;
  }

  String getErrorMessage() {
    return fieldController.state.errors.isNotEmpty ? fieldController.state.errors[0] : "";
  }

  String getForbiddenCharactersError([bool named = false]) {
    var message = "";
    if (containsForbiddenCharacters()) {
      message = "Remova os seguintes caracteres: [";
      var forbbidenCharacters = getForbbidenCharacters();
      var length = forbbidenCharacters.length;
      for (var i = 0; i < length; i++) {
        var char = forbbidenCharacters[i];

        if (i == 0 && length == 1) {
          message = "$message $char ].";
        } else if (i == 0) {
          message = "$message $char";
        } else if (i + 1 == length) {
          message = "$message, $char ].";
        } else {
          message = "$message, $char";
        }
      }
      var label = widget.decoration?.labelText;
      if (label != null && named) {
        message = "$label: $message";
      }
    }
    return message;
  }

  List<String> getForbbidenCharacters() {
    var list = <String>[];
    if (widget.forbiddenCharacters != null) {
      for (var match in widget.forbiddenCharacters!.allMatches(getValue())) {
        var char = match[0].toString();
        list.add(char.isNotEmpty ? char : "*espaço em branco");
      }
    }
    if (widget.forbidSpaces && getValue().contains(' ')) {
      list.add("espaço em branco");
    }
    return list;
  }

  void onValidate() {
    validationPipeline();
  }

  void onSet() => setValue();

  void onReset() {
    pipeline("");
  }

  void onChanged(String value) {
    pipeline(value);
  }

  // PIPELINE
  void pipeline(String value) {
    fieldController.setValue(value);
    valuePipeline();
    validationPipeline();
    changeState();
    if (fieldController.state.errors.isEmpty) {
      setValue();
    }
  }

  void valuePipeline() {
    widget.onChanged?.call(fieldController);
  }

  void validationPipeline() {
    fieldController.state.errors.clear();
    widget.onChanged?.call(fieldController);
    validate();
  }

  void changeState() {
    setState(() {
      errorMessage = getErrorMessage();
      controller.text = fieldController.state.value;
      controller.selection = TextSelection.fromPosition(TextPosition(offset: controller.text.length));
    });
  }

  @override
  Widget build(BuildContext context) {
    return Container(
      margin: widget.margin,
      child: TextField(
        controller: controller,
        enabled: widget.enabled,
        style: widget.style ?? defaultTextStyle(),
        decoration: inputDecoration(),
        onChanged: onChanged,
      ),
    );
  }

  InputDecoration inputDecoration() {
    var decoration = defaultInputDecoration();
    if (widget.decoration != null) {
      decoration = decoration.combineWith(widget.decoration!);
    }
    if (errorMessage.isEmpty) {
      return decoration;
    } else {
      return decoration.copyWith(
        errorText: errorMessage,
      );
    }
  }

  TextStyle defaultTextStyle() {
    return widget.formController.viewModel.styleProvider.auto(FontType.onBackground);
  }

  InputDecoration defaultInputDecoration() {
    return InputDecoration(
      floatingLabelBehavior: FloatingLabelBehavior.auto,
      labelStyle: errorMessage.isNotEmpty
          ? widget.formController.viewModel.styleProvider.auto(FontType.onError, FontSize.label)
          : widget.formController.viewModel.styleProvider.auto(FontType.onBackground, FontSize.label),
      hintText: !widget.enabled ? "Desabilitado" : null,
      hintStyle: widget.formController.viewModel.styleProvider.auto(FontType.onBackground),
      errorStyle: widget.formController.viewModel.styleProvider.auto(FontType.onError, FontSize.label),
      disabledBorder: OutlineInputBorder(
        borderSide: BorderSide(
          color: widget.formController.viewModel.colors.secondary,
        ),
      ),
    );
  }
}

class FormErrorsDisplay extends StatefulWidget {
  const FormErrorsDisplay({
    Key? key,
    required this.controller,
    this.padding,
    this.margin,
  }) : super(key: key);

  final FormController controller;
  final EdgeInsets? margin;
  final EdgeInsets? padding;

  @override
  State<FormErrorsDisplay> createState() => _FormErrorsDisplayState();
}

class _FormErrorsDisplayState extends State<FormErrorsDisplay> {
  List<String> errors = [];

  @override
  void initState() {
    super.initState();
    widget.controller.state.subscribe(onStateChanged);
  }

  @override
  void dispose() {
    widget.controller.state.unsubscribe(onStateChanged);
    super.dispose();
  }

  void onStateChanged(DynamicFormState state) {
    setState(() {
      errors = state.errors;
    });
  }

  @override
  Widget build(BuildContext context) {
    if (errors.isEmpty) {
      return Container();
    }
    return Container(
      margin: widget.margin,
      padding: widget.padding,
      child: Column(
        children: [
          Row(
            children: [
              Icon(
                Icons.warning_amber,
                color: widget.controller.viewModel.config.colorScheme.onError,
                size: widget.controller.viewModel.config.fontSize.data,
              ),
              const VerticalDivider(),
              Flexible(
                child: Text(
                  "Corrija os seguintes erros antes de tentar novamente:",
                  maxLines: 2,
                  style: widget.controller.viewModel.styleProvider.auto(FontType.onError),
                ),
              ),
            ],
          ),
          Text(
            getErrorText(),
            maxLines: 50,
            style: widget.controller.viewModel.styleProvider.auto(FontType.onError),
          ),
        ],
      ),
    );
  }

  String getErrorText() {
    var str = "";
    for (var error in errors) {
      str = '$str\n- $error';
    }
    return str;
  }
}

class IntField extends StatelessWidget {
  const IntField({
    Key? key,
    required this.formController,
    required this.getter,
    required this.setter,
    this.enabled = true,
    this.margin,
    this.style,
    this.decoration,
    this.forbiddenCharacters,
  }) : super(key: key);

  final FormController formController;
  final int Function() getter;
  final void Function(int data) setter;
  // optional
  final bool enabled;
  final EdgeInsets? margin;
  final TextStyle? style;
  final InputDecoration? decoration;
  final RegExp? forbiddenCharacters;

  String intGetter() {
    return getter().toString();
  }

  void intSetter(String value) {
    setter(int.parse(value));
  }

  @override
  Widget build(BuildContext context) {
    return DynamicFormField(
      formController: formController,
      getter: intGetter,
      setter: intSetter,
      enabled: enabled,
      forbidSpaces: true,
      validateOnChanged: true,
      margin: margin,
      style: style,
      decoration: decoration,
      forbiddenCharacters: forbiddenCharacters,
      onChanged: onChanged,
    );
  }

  void onChanged(FieldController controller) {
    var value = controller.state.value;
    value = value.replaceAll(RegExp("[^0-9]"), "");
    if (value.isEmpty) {
      value = "0";
    }
    controller.setValue(value);
  }
}

class DoubleField extends StatelessWidget {
  const DoubleField({
    Key? key,
    required this.formController,
    required this.getter,
    required this.setter,
    this.enabled = true,
    this.margin,
    this.style,
    this.decoration,
    this.forbiddenCharacters,
  }) : super(key: key);

  final FormController formController;
  final double Function() getter;
  final void Function(double data) setter;
  // optional
  final bool enabled;
  final EdgeInsets? margin;
  final TextStyle? style;
  final InputDecoration? decoration;
  final RegExp? forbiddenCharacters;

  String doubleGetter() {
    return getter().toString();
  }

  void doubleSetter(String value) {
    setter(double.parse(value));
  }

  @override
  Widget build(BuildContext context) {
    return DynamicFormField(
      formController: formController,
      getter: doubleGetter,
      setter: doubleSetter,
      enabled: enabled,
      forbidSpaces: true,
      validateOnChanged: true,
      margin: margin,
      style: style,
      decoration: decoration,
      forbiddenCharacters: forbiddenCharacters,
      onChanged: onChanged,
    );
  }

  void onChanged(FieldController controller) {
    var value = controller.state.value;
    value = value.replaceAll(RegExp("[^0-9]"), "");
    if (value.isEmpty) {
      value = "0";
    }
    controller.setValue(value);
  }
}

class CurrencyField extends StatefulWidget {
  const CurrencyField({
    Key? key,
    required this.formController,
    required this.getter,
    required this.setter,
    this.enabled = true,
    this.margin,
    this.style,
    this.decoration,
    this.forbiddenCharacters,
  }) : super(key: key);

  final FormController formController;
  final double Function() getter;
  final void Function(double data) setter;
  final int decimals = 2;
  // optional
  final bool enabled;
  final EdgeInsets? margin;
  final TextStyle? style;
  final InputDecoration? decoration;
  final RegExp? forbiddenCharacters;

  @override
  State<CurrencyField> createState() => _CurrencyFieldState();
}

class _CurrencyFieldState extends State<CurrencyField> {
  late CurrencyText text = CurrencyText(decimals: widget.decimals);

  void doubleSetter(String value) {
    widget.setter(double.parse(text.toDouble().toStringAsFixed(2)));
  }

  String doubleGetter() {
    return widget.getter().toString();
  }

  String normalize(String value) {
    value = value.replaceAll(RegExp("[^0-9]"), "");
    if (value.isEmpty) {
      value = "0";
    }
    return value;
  }

  List<String> trimZeroes(List<String> chars) {
    while (chars.isNotEmpty && chars[0] == '0') {
      chars.removeAt(0);
    }
    return chars;
  }

  void onChanged(FieldController controller) {
    var numbers = normalize(controller.state.value);
    var chars = numbers.characters.toList(); // [0, 0, 0, 9]
    var minCharCount = widget.decimals + 1;
    chars = trimZeroes(chars); // [9]
    text.clear();
    while (chars.length < minCharCount) {
      chars.insert(0, '0');
    }
    // chars = [0, 0, 9]
    for (var index = 0; index < widget.decimals; index++) {
      text.addNumber(
        Number(
          decimalPosition: -(widget.decimals - index),
          value: int.parse(chars.reversed.elementAt(index)),
        ),
      );
    }
    // chars = [1, 5, 0,| 0 ,0]
    for (var index = 0; index < chars.length - widget.decimals; index++) {
      text.addNumber(
        Number(
          decimalPosition: chars.length - widget.decimals - index,
          value: int.parse(chars.elementAt(index)),
        ),
      );
    }
    controller.setValue(text.toString());
  }

  @override
  Widget build(BuildContext context) {
    return DynamicFormField(
      formController: widget.formController,
      getter: doubleGetter,
      setter: doubleSetter,
      enabled: widget.enabled,
      forbidSpaces: true,
      validateOnChanged: true,
      margin: widget.margin,
      style: widget.style,
      decoration: widget.decoration,
      forbiddenCharacters: widget.forbiddenCharacters,
      onChanged: onChanged,
    );
  }
}

///  ///    ////  / //  / //  / / /

class FormStringField extends StatefulWidget {
  const FormStringField({
    Key? key,
    required this.getValue,
    required this.setValue,
    required this.formController,
    this.enabled = true,
    this.forbidSpaces = false,
    this.hintText,
    this.margin,
    this.style,
    this.decoration,
    this.forbiddenCharacters,
    this.fieldController,
  }) : super(key: key);

  final String Function() getValue;
  final void Function(String data) setValue;
  final FormControllerOld formController;
  // optional
  final bool enabled;
  final bool forbidSpaces;
  final String? hintText;
  final EdgeInsets? margin;
  final TextStyle? style;
  final InputDecoration? decoration;
  final RegExp? forbiddenCharacters;
  final FieldController? fieldController;

  @override
  State<FormStringField> createState() {
    return _FormStringFieldState();
  }
}

class _FormStringFieldState extends State<FormStringField> {
  late String errorMessage = "";
  late TextEditingController controller = TextEditingController();

  @override
  void initState() {
    super.initState();
    subscribeToEvents();
    controller.text = widget.getValue.call();
  }

  @override
  void dispose() {
    unsubscribeToEvents();
    super.dispose();
  }

  void subscribeToEvents() {
    widget.formController.submitEvent.subscribe(onSubmit);
    widget.formController.resetEvent.subscribe(onReset);
    if (widget.fieldController != null) {
      widget.fieldController!.resetEvent.subscribe(onReset);
      widget.fieldController!.getValueEvent.subscribe(onGetValue);
      widget.fieldController!.getIsValidEvent.subscribe(onGetIsValid);
    }
  }

  void unsubscribeToEvents() {
    widget.formController.submitEvent.unsubscribe(onSubmit);
    widget.formController.resetEvent.unsubscribe(onReset);
    if (widget.fieldController != null) {
      widget.fieldController!.resetEvent.unsubscribe(onReset);
      widget.fieldController!.getValueEvent.unsubscribe(onGetValue);
      widget.fieldController!.getIsValidEvent.unsubscribe(onGetIsValid);
    }
  }

  String getFieldValue() {
    return controller.text.trim();
  }

  void showError(String message) {
    setState(() {
      errorMessage = message;
    });
  }

  void hideError() {
    setState(() {
      errorMessage = "";
    });
  }

  bool getIsInputValid() {
    if (widget.forbiddenCharacters!.hasMatch(getFieldValue()) || getFieldValue().contains(" ") && widget.forbidSpaces) {
      return false;
    }
    return true;
  }

  List<String> getForbbidenCharacters() {
    var list = <String>[];
    for (var match in widget.forbiddenCharacters!.allMatches(getFieldValue())) {
      var char = match[0].toString();
      list.add(char.isNotEmpty ? char : "*espaço em branco");
    }
    if (RegExp(" ").allMatches(getFieldValue()).isNotEmpty) {
      list.add("espaços em branco");
    }
    return list;
  }

  String getErrorMessage(bool withName) {
    var message = "";
    if (widget.forbiddenCharacters!.hasMatch(getFieldValue()) || getFieldValue().contains(" ") && widget.forbidSpaces) {
      message = "Caracteres proibidos foram encontrados, remova os seguintes caracteres: [";
      var forbbidenCharacters = getForbbidenCharacters();
      var length = forbbidenCharacters.length;
      for (var i = 0; i < length; i++) {
        var char = forbbidenCharacters[i];

        if (i == 0 && length == 1) {
          message = "$message $char ].";
        } else if (i == 0) {
          message = "$message $char";
        } else if (i + 1 == length) {
          message = "$message, $char ].";
        } else {
          message = "$message, $char";
        }
      }
    } else {
      message = "Erro desconhecido.";
    }
    if (withName) {
      message = "${widget.hintText}: $message";
    }
    return message;
  }

  void onChanged(String value) {
    if (getIsInputValid()) {
      hideError();
      widget.setValue(getFieldValue());
    } else {
      showError(getErrorMessage(false));
    }
  }

  void onSubmit() {
    if (!getIsInputValid()) {
      showError(getErrorMessage(false));
      widget.formController.validationErrorEvent.invoke(getErrorMessage(true));
    } else {
      hideError();
      widget.setValue(getFieldValue());
    }
  }

  void onReset() {
    controller.clear();
  }

  String onGetValue() {
    return getFieldValue();
  }

  bool onGetIsValid() {
    return getIsInputValid();
  }

  @override
  Widget build(BuildContext context) {
    return Container(
      margin: widget.margin,
      child: TextField(
        controller: controller,
        enabled: widget.enabled,
        style: widget.style ?? defaultTextStyle(),
        decoration: inputDecoration(),
        onChanged: onChanged,
      ),
    );
  }

  InputDecoration inputDecoration() {
    var decoration = widget.decoration ?? defaultInputDecoration();
    if (errorMessage.isEmpty) {
      return decoration;
    } else {
      return decoration.copyWith(
        helperText: errorMessage,
        helperStyle: widget.formController.viewModel.styleProvider.auto(FontType.onError),
      );
    }
  }

  TextStyle defaultTextStyle() {
    return widget.formController.viewModel.styleProvider.auto(FontType.onBackground);
  }

  InputDecoration defaultInputDecoration() {
    return InputDecoration(
      labelText: widget.hintText,
      floatingLabelBehavior: FloatingLabelBehavior.always,
      labelStyle: errorMessage.isNotEmpty
          ? widget.formController.viewModel.styleProvider.auto(FontType.onError, FontSize.label)
          : widget.formController.viewModel.styleProvider.auto(FontType.onBackground, FontSize.label),
      hintText: !widget.enabled ? "Desabilitado" : null,
      //hintStyle: TextStyle(fontStyle: FontStyle.italic),

      focusedBorder: OutlineInputBorder(
        borderSide: BorderSide(
          color: errorMessage.isNotEmpty
              ? widget.formController.viewModel.config.colorScheme.onError
              : widget.formController.viewModel.config.colorScheme.onBackground,
        ),
      ),
      enabledBorder: OutlineInputBorder(
        borderSide: BorderSide(
          color: errorMessage.isNotEmpty
              ? widget.formController.viewModel.config.colorScheme.onError
              : widget.formController.viewModel.config.colorScheme.onBackground,
        ),
      ),
      disabledBorder: const OutlineInputBorder(
        borderSide: BorderSide(
          color: Colors.orange,
        ),
      ),
    );
  }
}
