import 'package:flutter/material.dart';
import 'package:modular_system_core/mscore.dart';

enum FieldType {
  textField,
  checkbox,
  dropdown,
  intNumber,
  doubleNumber,
  currency,
}

class NewEntityForm extends StatefulWidget {
  // ignore: use_key_in_widget_constructors
  const NewEntityForm({
    required this.title,
    required this.submitEvent,
    required this.errorEvent,
    this.commandIssuedEvent,
    this.fields = const [],
    this.children = const <Widget>[],
  });

  final String title;
  final AppEvent submitEvent;
  final AppEvent errorEvent;
  final AppEvent? commandIssuedEvent;
  final List<EntityFormFieldBuilder> fields;
  final List<Widget> children;

  @override
  State<StatefulWidget> createState() {
    // ignore: no_logic_in_create_state
    return _NewEntityFormState(
      title: title,
      fields: fields,
      children: children,
      submitEvent: submitEvent,
      commandIssuedEvent: commandIssuedEvent,
      errorEvent: errorEvent,
    );
  }
}

class _NewEntityFormState extends State<NewEntityForm> {
  _NewEntityFormState({
    required this.title,
    required this.fields,
    required this.submitEvent,
    required this.errorEvent,
    required this.commandIssuedEvent,
    this.children = const <Widget>[],
  }) {
    injectDependencies();
  }

  @override
  void dispose() {
    super.dispose();
  }

  String title;
  String errorMessage = "";
  AppEvent submitEvent;
  AppEvent doneEvent = AppEvent();
  AppEvent stateChangedEvent = AppEvent();
  AppEvent errorEvent;
  AppEvent? commandIssuedEvent;
  List<EntityFormFieldBuilder> fields;
  List<Widget> children;
  List<String> errors = [];

  @override
  Widget build(BuildContext context) {
    try {
      return Container(
        padding: const EdgeInsets.all(8),
        child: ListView(
          children: [
            Container(
              margin: const EdgeInsets.only(bottom: 50),
              child: Text(
                title,
                style: Theme.of(context).textTheme.headline2,
              ),
            ),
            Column(
              crossAxisAlignment: CrossAxisAlignment.start,
              children: children,
            ),
            Container(
              margin: const EdgeInsets.only(top: 100),
              child: Text(
                errorMessage,
                style: TextStyle(
                  color: Colors.red,
                  fontSize: Theme.of(context).textTheme.headline3?.fontSize,
                ),
              ),
            ),
            OutlinedButton(
              child: Row(
                children: [
                  Padding(
                    padding: const EdgeInsets.only(right: 10),
                    child: Icon(
                      Icons.library_add,
                      color: Theme.of(context).textTheme.headline4?.color,
                    ),
                  ),
                  Text(
                    "Criar",
                    style: Theme.of(context).textTheme.headline4,
                  ),
                ],
              ),
              style: OutlinedButton.styleFrom(
                padding: const EdgeInsets.all(20),
              ),
              onPressed: () => onButtonPressed(),
            ),
          ],
        ),
      );
    } catch (e) {
      errorEvent.invoke(e);
      throw e;
    }
  }

  void onButtonPressed() {
    try {
      errors.clear();
      doneEvent.invoke();
      setErrorMessage();
      if (errors.isEmpty) {
        submitEvent.invoke();
      }
    } catch (e) {
      errorEvent.invoke(e);
    }
  }

  void injectDependencies() {
    var count = 1;
    for (var i = 0; i < children.length; i++) {
      var widget = children[i];
      if (widget is EntityFormField) {
        inject(i, count, widget);
        count++;
      } else if (widget is MultiChildRenderObjectWidget) {
        for (var _widget in widget.children) {
          if (_widget is EntityFormField) {
            inject(i, count, _widget);
            count++;
          }
        }
      }
    }
  }

  void inject(int i, int count, EntityFormField widget) {
    var field = EntityFormField(
      // dependency injection
      fieldId: count,
      doneEvent: doneEvent,
      stateChangedEvent: stateChangedEvent,
      commandIssuedEvent: commandIssuedEvent,
      addError: addError,
      // reinjection of original values
      hintText: widget.hintText,
      type: widget.type,
      getValue: widget.getValue,
      setValue: widget.setValue,
      forbidSpaces: widget.forbidSpaces,
      enabled: widget.enabled,
      dropdownItems: widget.dropdownItems,
      forbiddenCharacters: widget.forbiddenCharacters,
      stateChangedListner: widget.stateChangedListner,
      commandIssuedListner: widget.commandIssuedListner,
      decimalPointCount: widget.decimalPointCount,
      numberSymbol: widget.numberSymbol,
      trailingNumberSymbol: widget.trailingNumberSymbol,
    );
    children.removeAt(i);
    children.insert(i, field);
  }

  void addError(String message) {
    errors.add(message);
  }

  void setErrorMessage() {
    var message = "";
    if (errors.isNotEmpty) {
      message = "Erros foram encontrados, corrigir antes de tentar de novo:\n";
      for (var err in errors) {
        message = "$message \n - $err\n";
      }
    }
    setState(() {
      errorMessage = message;
    });
  }
}

class EntityFormFieldBuilder {
  EntityFormFieldBuilder({
    required this.setValue,
    required this.type,
    required this.hintText,
    this.forbidSpaces = false,
    this.enabled = true,
    this.customEvent,
    this.shouldEnable,
    this.forbbidenCharacters,
    this.style,
    this.inputDecoration,
    this.dropdownItems,
    this.stateChangedListner,
  });

  final AppEvent? customEvent;
  final Function(dynamic value) setValue;
  final Function(Map<int, dynamic> data)? stateChangedListner;
  final bool Function()? shouldEnable;
  final bool forbidSpaces;
  final bool enabled;
  final FieldType type;
  final String hintText;
  final String? forbbidenCharacters;
  final TextStyle? style;
  final InputDecoration? inputDecoration;
  final Map<int, String>? dropdownItems;
}

class EntityFormField extends StatefulWidget {
  // ignore: use_key_in_widget_constructors
  const EntityFormField({
    required this.hintText,
    required this.type,
    required this.getValue,
    required this.setValue,
    this.fieldId = -1,
    this.doneEvent,
    this.stateChangedEvent,
    this.commandIssuedEvent,
    this.addError,
    this.forbidSpaces = true,
    this.enabled = true,
    this.stateChangedListner,
    this.commandIssuedListner,
    this.dropdownItems,
    this.style,
    this.inputDecoration,
    this.forbiddenCharacters,
    this.decimalPointCount = 2,
    this.numberSymbol = "",
    this.trailingNumberSymbol = false,
  });

  static String nothingRegex = 'a^';
  static String onlyNumbersAndLettersAllowedRegex = '[^a-z, A-Z, 0-9]';

  final int fieldId;
  final AppEvent? doneEvent;
  final AppEvent? stateChangedEvent;
  final AppEvent? commandIssuedEvent;
  final FieldType type;
  final bool forbidSpaces;
  final bool enabled;
  final void Function(dynamic value) setValue;
  final dynamic Function() getValue;
  final void Function(String message)? addError;
  final void Function(Map<int, dynamic> data)? stateChangedListner;
  final void Function(_EntityFormFieldState state)? commandIssuedListner;

  final bool trailingNumberSymbol;
  final String hintText;
  final String? forbiddenCharacters;
  final String numberSymbol;
  final TextStyle? style;
  final InputDecoration? inputDecoration;
  final Map<int, String>? dropdownItems;
  final int decimalPointCount;

  @override
  State<StatefulWidget> createState() {
    // ignore: no_logic_in_create_state
    return _EntityFormFieldState(
      doneEvent: doneEvent,
      getValue: getValue,
      setValue: setValue,
      type: type,
      addError: addError,
      hintText: hintText,
      style: style,
      inputDecoration: inputDecoration,
      forbiddenCharacters: forbiddenCharacters,
      forbidSpaces: forbidSpaces,
      dropdownItems: dropdownItems,
      enabled: enabled,
      stateChangedEvent: stateChangedEvent,
      fieldId: fieldId,
      stateChangedListner: stateChangedListner,
      commandIssuedEvent: commandIssuedEvent,
      commandIssuedListner: commandIssuedListner,
      decimalPointCount: decimalPointCount,
      numberSymbol: numberSymbol,
      trailingNumberSymbol: trailingNumberSymbol,
    );
  }
}

class _EntityFormFieldState extends State<EntityFormField> {
  _EntityFormFieldState({
    required this.hintText,
    required this.type,
    required this.getValue,
    required this.setValue,
    required this.fieldId,
    this.numberSymbol = "",
    this.trailingNumberSymbol = false,
    this.forbidSpaces = true,
    this.enabled = true,
    this.decimalPointCount = 2,
    this.doneEvent,
    this.stateChangedEvent,
    this.commandIssuedEvent,
    this.addError,
    this.stateChangedListner,
    this.commandIssuedListner,
    this.dropdownItems,
    this.style,
    this.inputDecoration,
    String? forbiddenCharacters,
  }) {
    doneEvent!.subscribe(onDone);
    if (commandIssuedListner != null) {
      commandIssuedEvent!.subscribe(onCommandIssued);
    }
    if (stateChangedListner != null) {
      stateChangedEvent!.subscribe(onStateChanged);
    }
    setInitialValue();
    forbiddenCharactersRegex = RegExp(forbiddenCharacters ??
        EntityFormField.onlyNumbersAndLettersAllowedRegex);
  }

  @override
  void dispose() {
    doneEvent!.unsubscribe(onDone);
    if (commandIssuedListner != null) {
      commandIssuedEvent!.unsubscribe(onCommandIssued);
    }
    if (stateChangedListner != null) {
      stateChangedEvent!.unsubscribe(onStateChanged);
    }
    super.dispose();
  }

  bool forbidSpaces;
  int fieldId;
  FieldType type;
  AppEvent? doneEvent;
  AppEvent? stateChangedEvent;
  AppEvent? commandIssuedEvent;
  late RegExp forbiddenCharactersRegex;
  TextEditingController controller = TextEditingController();
  void Function(dynamic value) setValue;
  dynamic Function() getValue;
  void Function(String message)? addError;
  void Function(Map<int, dynamic> data)? stateChangedListner;
  void Function(_EntityFormFieldState data)? commandIssuedListner;
  // looks stuff
  String hintText;
  TextStyle? style;
  InputDecoration? inputDecoration;
  // inner state
  bool enabled;
  bool checkboxValue = false;
  bool trailingNumberSymbol = false;
  String? errorMessage;
  int dropdownValue = 0;
  Map<int, String>? dropdownItems;
  String numberSymbol;
  int decimalPointCount;
  late CurrencyText doubleValue = CurrencyText(
    symbol: "",
    decimalPointCount: decimalPointCount,
  );
  CurrencyText currencyText = CurrencyText(
    symbol: "R\$",
  );

  @override
  Widget build(BuildContext context) {
    if (type == FieldType.textField) {
      return getTextField(context);
    } else if (type == FieldType.dropdown) {
      return getDropdown(context);
    } else if (type == FieldType.checkbox) {
      return getCheckbox(context);
    } else if (type == FieldType.intNumber) {
      return getIntField(context);
    } else if (type == FieldType.doubleNumber) {
      return getDoubleField(context);
    } else if (type == FieldType.currency) {
      return getCurrencyField(context);
    } else {
      throw AppException("");
    }
  }

  setInitialValue() {
    var value = getValue.call();
    if (value == null) {
      return;
    }
    if (type == FieldType.textField) {
      controller.text = value;
    } else if (type == FieldType.dropdown) {
      dropdownValue = value;
    } else if (type == FieldType.checkbox) {
      checkboxValue = value;
    } else if (type == FieldType.intNumber) {
      controller.text = value.toString();
      onChanged(value.toString());
    } else if (type == FieldType.doubleNumber) {
      controller.text = value.toString();
      onChanged(value.toString());
    } else if (type == FieldType.currency) {
      controller.text = value.toString();
      onChanged(value.toString());
    }
  }

  // BUILD BODIES
  Widget getTextField(BuildContext context) {
    return Container(
      margin: const EdgeInsets.symmetric(
        vertical: 20,
      ),
      child: TextField(
        controller: controller,
        enabled: enabled,
        style: style ?? defaultStyle(context),
        decoration:
            inputDecoration ?? defaultInputDecoration(hintText, context),
        onChanged: onChanged,
      ),
    );
  }

  Widget getDropdown(BuildContext context) {
    return Container(
      margin: const EdgeInsets.symmetric(
        vertical: 20,
      ),
      child: Column(
        crossAxisAlignment: CrossAxisAlignment.start,
        children: [
          Text(
            hintText,
            style: const TextStyle(
              color: Colors.black,
            ),
          ),
          DropdownButtonHideUnderline(
            child: DropdownButton(
              isExpanded: true,
              value: dropdownValue,
              focusNode: FocusNode(canRequestFocus: false),
              items: getDropdownItemsList(),
              onChanged: (value) {
                setState(() {
                  dropdownValue = value as int;

                  onDone(value);
                });
              },
              style: TextStyle(
                color: Colors.black,
                fontSize: Theme.of(context).textTheme.headline4?.fontSize,
              ),
              hint: Text(
                hintText,
                style: TextStyle(
                  color: Colors.black,
                  fontSize: Theme.of(context).textTheme.headline3?.fontSize,
                ),
              ),
            ),
          ),
        ],
      ),
    );
  }

  List<DropdownMenuItem<int>> getDropdownItemsList() {
    var list = <DropdownMenuItem<int>>[];
    for (var i = 0; i < dropdownItems!.length; i++) {
      var item = dropdownItems![i];
      list.add(
        DropdownMenuItem(
          child: Text(item!),
          value: i,
        ),
      );
    }

    return list;
  }

  Widget getCheckbox(BuildContext context) {
    return Container(
      margin: const EdgeInsets.symmetric(
        vertical: 20,
      ),
      child: Column(
        crossAxisAlignment: CrossAxisAlignment.start,
        children: [
          Text(hintText),
          Checkbox(
            value: checkboxValue,
            onChanged: (value) {
              setState(() {
                checkboxValue = value!;
                onDone(value);
              });
            },
          ),
        ],
      ),
    );
  }

  Widget getIntField(BuildContext context) {
    return Container(
      margin: const EdgeInsets.symmetric(
        vertical: 20,
      ),
      child: TextField(
        controller: controller,
        enabled: enabled,
        style: style ?? defaultStyle(context),
        decoration:
            inputDecoration ?? defaultInputDecoration(hintText, context),
        onChanged: onChanged,
      ),
    );
  }

  Widget getDoubleField(BuildContext context) {
    return Container(
      margin: const EdgeInsets.symmetric(
        vertical: 20,
      ),
      child: TextField(
        controller: controller,
        enabled: enabled,
        style: style ?? defaultStyle(context),
        decoration:
            inputDecoration ?? defaultInputDecoration(hintText, context),
        onChanged: onChanged,
      ),
    );
  }

  Widget getCurrencyField(BuildContext context) {
    return Container(
      margin: const EdgeInsets.symmetric(
        vertical: 20,
      ),
      child: TextField(
        controller: controller,
        enabled: enabled,
        style: style ?? defaultStyle(context),
        decoration:
            inputDecoration ?? defaultInputDecoration(hintText, context),
        onChanged: onChanged,
      ),
    );
  }

  // TEXT STYLE
  TextStyle? defaultStyle(BuildContext context) {
    if (type == FieldType.textField) {
      return defaultTextInputStyle();
    } else {
      return Theme.of(context).textTheme.headline4;
    }
  }

  TextStyle defaultTextInputStyle() {
    return TextStyle(
      color: Colors.black,
      fontSize: Theme.of(context).textTheme.headline4?.fontSize,
    );
  }

  // INPUT BORDER
  InputDecoration defaultInputDecoration(
      String hintText, BuildContext context) {
    if (type == FieldType.textField) {
      return defaultTextInputDecoration(hintText, context);
    } else if (type == FieldType.intNumber) {
      return defaultNumberBorder(hintText, context);
    } else if (type == FieldType.doubleNumber) {
      return defaultNumberBorder(hintText, context);
    } else if (type == FieldType.currency) {
      return defaultCurrencyBorder(hintText, context);
    }
    //
    return defaultTextInputDecoration(hintText, context);
  }

  InputDecoration defaultTextInputDecoration(
      String hintText, BuildContext context) {
    return InputDecoration(
      labelText: hintText,
      floatingLabelBehavior: FloatingLabelBehavior.always,
      labelStyle: TextStyle(
        color: errorMessage != null ? Colors.red : Colors.black,
        fontSize: Theme.of(context).textTheme.headline3?.fontSize,
      ),
      hintText: !enabled ? "Desabilitado" : null,
      //hintStyle: TextStyle(fontStyle: FontStyle.italic),
      helperText: errorMessage,
      helperStyle: TextStyle(
        color: Colors.red,
        fontSize: Theme.of(context).textTheme.headline4?.fontSize,
      ),
      focusedBorder: OutlineInputBorder(
        borderSide: BorderSide(
          color: errorMessage != null ? Colors.red : Colors.black,
        ),
      ),
      enabledBorder: OutlineInputBorder(
        borderSide: BorderSide(
          color: errorMessage != null ? Colors.red : Colors.black,
        ),
      ),
      disabledBorder: const OutlineInputBorder(
        borderSide: BorderSide(
          color: Colors.orange,
        ),
      ),
    );
  }

  InputDecoration defaultNumberBorder(String hintText, BuildContext context) {
    return defaultTextInputDecoration(hintText, context).copyWith(
      hintText: "Inserir números",
    );
  }

  InputDecoration defaultCurrencyBorder(String hintText, BuildContext context) {
    return defaultTextInputDecoration(hintText, context).copyWith(
      hintText: "R\$ 0,00",
    );
  }

// ON CHANGED
  void onChanged(dynamic data) {
    if (type == FieldType.textField) {
      textFieldOnChanged(data);
    } else if (type == FieldType.intNumber) {
      intOnChanged(data);
    } else if (type == FieldType.doubleNumber) {
      doubleOnChanged(data);
    } else if (type == FieldType.currency) {
      currencyOnChanged(data);
    }
  }

  void textFieldOnChanged(String data) {
    if (getIsInputValid()) {
      hideError();
    } else {
      showError(getErrorMessage(false));
    }
  }

  void intOnChanged(String data) {
    var number = getCurrentValue();
    if (numberSymbol.isNotEmpty) {
      if (trailingNumberSymbol) {
        controller.text = "$number $numberSymbol";
      } else {
        controller.text = "$numberSymbol $number";
      }
    } else {
      controller.text = "$number";
    }
    controller.selection = TextSelection.fromPosition(
        TextPosition(offset: controller.text.length));
  }

  void doubleOnChanged(String dataa) {
    var value = getCurrentValue();
    doubleValue.clear();
    for (var i = 0; i < value.length; i++) {
      var char = value.substring(i, i + 1);
      var number = int.parse(char);
      doubleValue.addNumber(number);
    }
    if (numberSymbol.isNotEmpty) {
      if (trailingNumberSymbol) {
        controller.text = "${doubleValue.toString()} $numberSymbol";
      } else {
        controller.text = "$numberSymbol ${doubleValue.toString()}";
      }
    } else {
      controller.text = doubleValue.toString();
    }
    if (trailingNumberSymbol) {
      controller.selection = TextSelection.fromPosition(TextPosition(
          offset: controller.text.length - (numberSymbol.length + 1)));
    } else {
      controller.selection = TextSelection.fromPosition(
          TextPosition(offset: controller.text.length));
    }
  }

  void currencyOnChanged(String data) {
    var value = getCurrentValue();
    currencyText.clear();
    for (var i = 0; i < value.length; i++) {
      var char = value.substring(i, i + 1);
      var number = int.parse(char);
      currencyText.addNumber(number);
    }
    controller.text = currencyText.toCurrencyString();
    controller.selection = TextSelection.fromPosition(
        TextPosition(offset: controller.text.length));
  }

  // ON DONE EVENT
  void onDone(dynamic data) {
    if (type == FieldType.textField) {
      textFieldOnDone(data);
    } else if (type == FieldType.dropdown) {
      dropdownOnDone(data);
    } else if (type == FieldType.checkbox) {
      checkboxOnDone(data);
    } else if (type == FieldType.intNumber) {
      intOnDone(data);
    } else if (type == FieldType.doubleNumber) {
      doubleOnDone(data);
    } else if (type == FieldType.currency) {
      currencyOnDone(data);
    }
    invokeStateChanged();
  }

  void textFieldOnDone(String data) {
    if (getIsInputValid()) {
      hideError();
      setValue.call(getCurrentValue());
    } else {
      showError(getErrorMessage(false));
      addError!.call(getErrorMessage(true));
    }
  }

  void dropdownOnDone(int data) {
    setValue.call(getCurrentValue());
  }

  void checkboxOnDone(bool data) {
    setValue.call(getCurrentValue());
  }

  void intOnDone(int data) {
    setValue.call(getCurrentValue());
  }

  void doubleOnDone(double data) {
    setValue.call(getCurrentValue());
  }

  void currencyOnDone(int data) {
    setValue.call(currencyText.toDouble());
  }

  // STATE CHANGED EVENT
  void invokeStateChanged() {
    stateChangedEvent!.invoke({fieldId: getCurrentValue()});
  }

  void onStateChanged(dynamic data) {
    stateChangedListner!.call(data);
  }

  // COMMAND ISSUED EVENT (change the  event's name please)
  void onCommandIssued(dynamic id) {
    if (id == fieldId && commandIssuedListner != null) {
      commandIssuedListner!.call(this);
    }
  }

  // OTHER STUFF
  dynamic getCurrentValue() {
    if (type == FieldType.textField) {
      var value = controller.text;
      return value.trim();
    } else if (type == FieldType.dropdown) {
      return dropdownValue;
    } else if (type == FieldType.intNumber) {
      var number = controller.text.replaceAll(RegExp("[^0-9]"), "");
      return number.isNotEmpty ? int.parse(number) : 0;
    } else if (type == FieldType.doubleNumber) {
      return controller.text.replaceAll(RegExp("[^0-9]"), "");
    } else if (type == FieldType.currency) {
      return controller.text.replaceAll(RegExp("[^0-9]"), "");
    } else {
      return checkboxValue;
    }
  }

  bool getIsInputValid() {
    if (getCurrentValue().contains(forbiddenCharactersRegex) ||
        getCurrentValue().contains(" ") && forbidSpaces) {
      return false;
    }
    return true;
  }

  String getErrorMessage(bool withName) {
    var message = "";
    if (getCurrentValue().contains(forbiddenCharactersRegex) ||
        getCurrentValue().contains(" ") && forbidSpaces) {
      message =
          "Caracteres proibidos foram encontrados, remova os seguintes caracteres: [";
      var forbbidenCharacters = getForbbidenCharacters();
      var length = forbbidenCharacters.length;
      for (var i = 0; i < length; i++) {
        var char = forbbidenCharacters[i];

        if (i == 0 && length == 1) {
          message = "$message $char ].";
        } else if (i == 0) {
          message = "$message $char";
        } else if (i + 1 == length) {
          message = "$message, $char ].";
        } else {
          message = "$message, $char";
        }
      }
    } else {
      message = "Erro desconhecido.";
    }
    if (withName) {
      message = "$hintText: $message";
    }
    return message;
  }

  List<String> getForbbidenCharacters() {
    var list = <String>[];
    for (var match in forbiddenCharactersRegex.allMatches(getCurrentValue())) {
      var char = match[0].toString();
      list.add(char.isNotEmpty ? char : "*espaço em branco");
    }
    if (RegExp(" ").allMatches(getCurrentValue()).isNotEmpty) {
      list.add("espaços em branco");
    }
    return list;
  }

  void showError(String message) {
    setState(() {
      errorMessage = message;
    });
  }

  void hideError() {
    setState(() {
      errorMessage = null;
    });
  }
}

class CurrencyText {
  CurrencyText({
    required this.symbol,
    this.decimalPointCount = 2,
  });

  int decimalPointCount;
  String symbol;
  List<int> values = [];

  @override
  String toString() {
    var string = "";
    while (values.length < decimalPointCount + 1) {
      values.insert(0, 0);
    }
    for (var i = 0; i < values.length; i++) {
      var value = values.reversed.elementAt(i);
      if (i == decimalPointCount) {
        string = "$value,$string";
      } else {
        string = "$value$string";
      }
    }
    return string;
  }

  String toCurrencyString() {
    var string = toString();
    return "$symbol $string";
  }

  double toDouble() {
    var string = toString();
    string = string.replaceAll(",", ".");
    return double.parse(string);
  }

  void addNumber(int number) {
    if (number != 0 || values.where((element) => element != 0).isNotEmpty) {
      values.add(number);
    }
  }

  void clear() {
    values.clear();
  }
}
