import 'package:flutter/material.dart';
import 'package:modular_system_core/mscore.dart';

class EntityUpdateForm {
  EntityUpdateForm({
    required this.entity,
    required this.controller,
  });

  AppEntity entity;
  FormControllerOld controller;

  EntityTextUpdateField textUpdateField({
    bool? forbidSpaces,
    required String id,
    required String updateName,
    required String hintText,
    required String currentValue,
    RegExp? forbiddenCharacters,
  }) {
    return EntityTextUpdateField(
      forbidSpaces: forbidSpaces,
      id: id,
      updateName: updateName,
      hintText: hintText,
      currentValue: currentValue,
      formController: controller,
      entity: entity,
      forbiddenCharacters: forbiddenCharacters,
    );
  }
}

class UpdateForm {
  UpdateForm({
    required this.controller,
  });

  FormControllerOld controller;

  TextUpdateField textUpdateField({
    bool? forbidSpaces,
    required String hintText,
    required String currentValue,
    required Future<dynamic> Function(String, Event) submitCallback,
    RegExp? forbiddenCharacters,
  }) {
    return TextUpdateField(
      forbidSpaces: forbidSpaces,
      hintText: hintText,
      currentValue: currentValue,
      controller: controller,
      submitCallback: submitCallback,
      forbiddenCharacters: forbiddenCharacters,
    );
  }
}

class TextUpdateField extends StatefulWidget {
  // ignore: use_key_in_widget_constructors
  const TextUpdateField({
    this.forbidSpaces,
    required this.hintText,
    required this.currentValue,
    required this.controller,
    this.fieldController,
    required this.submitCallback,
    this.forbiddenCharacters,
  });

  final bool? forbidSpaces;
  final String hintText;
  final String currentValue;
  final FormControllerOld controller;
  final FieldController? fieldController;
  final Future Function(String value, Event success) submitCallback;
  final RegExp? forbiddenCharacters;

  @override
  State<TextUpdateField> createState() {
    // ignore: no_logic_in_create_state
    return _TextUpdateFieldState(
      hintText: hintText,
      currentValue: currentValue,
      formController: controller,
      fieldController: fieldController,
      submitCallback: submitCallback,
    );
  }
}

class _TextUpdateFieldState extends State<TextUpdateField> {
  _TextUpdateFieldState({
    this.forbidSpaces,
    required this.hintText,
    required this.currentValue,
    required this.formController,
    this.fieldController,
    required this.submitCallback,
    this.forbiddenCharacters,
  }) {
    subscribeToEvents();
    fieldController ??= FieldController();
  }

  // injected
  final bool? forbidSpaces;
  final String hintText;
  final FormControllerOld formController;
  late FieldController? fieldController;
  final RegExp? forbiddenCharacters;
  late AppForm form = AppForm(controller: formController);
  // state
  bool isEditMode = false;
  String currentValue;
  String editorValue = '';
  Event successEvent = Event();
  late Future submitFuture = Future(() {});
  Future Function(String value, Event success) submitCallback;

  @override
  void dispose() {
    unsubscribeToEvents();
    super.dispose();
  }

  void subscribeToEvents() {
    successEvent.subscribe(onSuccess);
  }

  void unsubscribeToEvents() {
    successEvent.unsubscribe(onSuccess);
  }

  @override
  Widget build(BuildContext context) {
    return ExpansionPanelList(
      children: [
        ExpansionPanel(
          isExpanded: isEditMode,
          canTapOnHeader: true,
          headerBuilder: (context, isExpanded) {
            return Padding(
              padding: const EdgeInsets.symmetric(horizontal: 15),
              child: Row(
                children: [
                  Icon(
                    Icons.edit,
                    size: formController.viewModel.config.fontSize.data,
                    color: formController.viewModel.config.colorScheme.onBackground,
                  ),
                  const VerticalDivider(),
                  Text(
                    "Valor Original: $currentValue",
                    style: formController.viewModel.styleProvider.auto(FontType.onBackground),
                  ),
                ],
              ),
            );
          },
          body: Container(
            padding: EdgeInsets.all(formController.viewModel.config.fontSize.data),
            decoration: BoxDecoration(
              border: Border(
                top: BorderSide.none,
                bottom: BorderSide(
                  color: formController.viewModel.config.colorScheme.onBackground,
                ),
                right: BorderSide(
                  color: formController.viewModel.config.colorScheme.onBackground,
                ),
                left: BorderSide(
                  color: formController.viewModel.config.colorScheme.onBackground,
                ),
              ),
            ),
            child: Column(
              children: [
                form.textField(
                  forbidSpaces: forbidSpaces ?? false,
                  hintText: hintText,
                  fieldController: fieldController,
                  forbiddenCharacters: forbiddenCharacters,
                  getValue: () => editorValue,
                  setValue: (value) => editorValue = value,
                ),
                Row(
                  mainAxisAlignment: MainAxisAlignment.end,
                  children: [
                    FutureBuilder(
                      future: submitFuture,
                      builder: (context, snapshot) {
                        if (snapshot.connectionState == ConnectionState.done) {
                          return OutlinedButton(
                            style: OutlinedButton.styleFrom(
                              padding: EdgeInsets.all(
                                formController.viewModel.config.fontSize.data,
                              ),
                            ),
                            child: Text(
                              "Atualizar",
                              style: formController.viewModel.styleProvider.auto(FontType.onBackground),
                            ),
                            onPressed: onSubmit,
                          );
                        }
                        return CircularProgressIndicator(
                          color: formController.viewModel.config.colorScheme.tertiary,
                        );
                      },
                    ),
                    VerticalDivider(
                      width: formController.viewModel.config.fontSize.data,
                    ),
                    OutlinedButton(
                      style: OutlinedButton.styleFrom(
                        padding: EdgeInsets.all(
                          formController.viewModel.config.fontSize.data,
                        ),
                      ),
                      child: Text(
                        "Cancelar",
                        style: formController.viewModel.styleProvider.auto(FontType.onBackground),
                      ),
                      onPressed: () {
                        setState(() {
                          isEditMode = false;
                        });
                        fieldController!.resetEvent.invoke();
                      },
                    ),
                  ],
                )
              ],
            ),
          ),
        ),
      ],
      expansionCallback: (index, isOpen) {
        setState(() {
          isEditMode = !isEditMode;
        });
      },
    );
  }

  void onSubmit() {
    setState(() {
      submitFuture = Future(() async {
        try {
          if (!formController.getIsFormValid()) {
            return;
          }
          await submitCallback(editorValue, successEvent);
        } catch (e) {
          formController.viewModel.errorHandler.handle(e);
        }
      });
    });
  }

  void onSuccess() {
    setState(() {
      currentValue = editorValue;
      isEditMode = false;
    });
    fieldController!.resetEvent.invoke();
  }
}

class EntityTextUpdateField extends StatelessWidget {
  // ignore: use_key_in_widget_constructors
  EntityTextUpdateField({
    this.forbidSpaces,
    required this.id,
    required this.updateName,
    required this.hintText,
    required this.currentValue,
    required this.formController,
    required this.entity,
    this.forbiddenCharacters,
  });

  final bool? forbidSpaces;
  final String id;
  final String updateName;
  final String hintText;
  final String currentValue;
  final FormControllerOld formController;
  final FieldController fieldController = FieldController();
  final AppEntity entity;
  final RegExp? forbiddenCharacters;

  @override
  Widget build(BuildContext context) {
    return TextUpdateField(
      forbidSpaces: forbidSpaces,
      hintText: hintText,
      currentValue: currentValue,
      controller: formController,
      fieldController: fieldController,
      submitCallback: submitAsync,
      forbiddenCharacters: forbiddenCharacters,
    );
  }

  Future<dynamic> submitAsync(String value, Event success) async {
    await entity.controller.updateAsync(getRequest());
    success.invoke();
  }

  WebUpdateRequest getRequest() {
    var req = WebUpdateRequest(id: id);
    var update = WebUpdateField();
    update.isActive = true;
    update.name = updateName;
    update.type = WebDataType.string;
    update.value = fieldController.getValueEvent.invoke();
    req.updates.add(update);
    return req;
  }
}

//

class EditorField extends StatefulWidget {
  // ignore: use_key_in_widget_constructors
  const EditorField({
    this.forbidSpaces,
    required this.initialValue,
    required this.hintText,
    required this.controller,
    required this.fieldController,
    required this.setValue,
    this.forbiddenCharacters,
    this.margin,
  });

  final bool? forbidSpaces;
  final String hintText;
  final String initialValue;
  final FormControllerOld controller;
  final FieldController fieldController;
  final void Function(String value) setValue;
  final RegExp? forbiddenCharacters;
  final EdgeInsets? margin;

  @override
  State<EditorField> createState() {
    // ignore: no_logic_in_create_state
    return _EditorFieldState(
      forbidSpaces: forbidSpaces,
      hintText: hintText,
      initialValue: initialValue,
      controller: controller,
      fieldController: fieldController,
      setValue: setValue,
      forbiddenCharacters: forbiddenCharacters,
      margin: margin,
    );
  }
}

class _EditorFieldState extends State<EditorField> {
  _EditorFieldState({
    this.forbidSpaces,
    required this.initialValue,
    required this.hintText,
    required this.controller,
    required this.fieldController,
    required this.setValue,
    this.forbiddenCharacters,
    this.margin,
  }) {}

  final bool? forbidSpaces;
  final String hintText;
  final FormControllerOld controller;
  final FieldController fieldController;
  final void Function(String value) setValue;
  final RegExp? forbiddenCharacters;
  final EdgeInsets? margin;
  bool isHover = false;
  bool isHoverConfirmBtn = false;
  bool isHoverCancelBtn = false;
  bool isEditMode = false;
  bool isUpdating = false;
  String initialValue;
  late String fieldValue = initialValue;
  late AppForm form = AppForm(controller: controller);

  @override
  void dispose() {
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return Container(
      margin: margin,
      child: isEditMode ? editor() : button(),
    );
  }

  Widget button() {
    return OutlinedButton(
      style: OutlinedButton.styleFrom(
        backgroundColor: controller.viewModel.config.colorScheme.primary,
        padding: const EdgeInsets.all(30),
        primary: controller.viewModel.config.colorScheme.tertiary,
        side: BorderSide(
          color: isHover ? controller.viewModel.config.colorScheme.onTertiaryContainer : controller.viewModel.config.colorScheme.onPrimary,
        ),
        shape: const RoundedRectangleBorder(
          borderRadius: BorderRadius.zero,
        ),
      ),
      child: Row(
        children: [
          Icon(
            Icons.edit,
            color: controller.viewModel.config.colorScheme.tertiary,
            size: controller.viewModel.config.fontSize.data,
          ),
          const VerticalDivider(),
          Flexible(
            child: Row(
              children: [
                Text(
                  '$hintText: ',
                  maxLines: 3,
                  softWrap: true,
                  style: controller.viewModel.styleProvider.auto(FontType.onTertiaryContainer),
                ),
                Text(
                  fieldValue,
                  maxLines: 3,
                  softWrap: true,
                  style: controller.viewModel.styleProvider.auto(FontType.onPrimary),
                ),
              ],
            ),
          ),
        ],
      ),
      onHover: (value) {
        setState(() {
          isHover = value;
        });
      },
      onPressed: () {
        setState(() {
          isEditMode = true;
        });
      },
    );
  }

  Widget editor() {
    return Container(
      padding: const EdgeInsets.all(7),
      decoration: BoxDecoration(
        border: Border.all(
          color: controller.viewModel.config.colorScheme.tertiary,
        ),
      ),
      child: Column(
        children: [
          form.textField(
            forbidSpaces: forbidSpaces ?? false,
            hintText: hintText,
            fieldController: fieldController,
            forbiddenCharacters: forbiddenCharacters,
            getValue: () => fieldValue,
            setValue: (value) {},
          ),
          Row(
            mainAxisAlignment: MainAxisAlignment.end,
            children: [
              OutlinedButton(
                style: OutlinedButton.styleFrom(
                  backgroundColor: controller.viewModel.config.colorScheme.primary,
                  primary: controller.viewModel.config.colorScheme.tertiary,
                  padding: EdgeInsets.all(
                    controller.viewModel.config.fontSize.data,
                  ),
                  side: BorderSide(
                    color: isHoverConfirmBtn
                        ? controller.viewModel.config.colorScheme.onTertiaryContainer
                        : controller.viewModel.config.colorScheme.onPrimary,
                  ),
                  shape: const RoundedRectangleBorder(
                    borderRadius: BorderRadius.zero,
                  ),
                ),
                child: Text(
                  "Confirmar",
                  style: isHoverConfirmBtn
                      ? controller.viewModel.styleProvider.auto(FontType.onTertiaryContainer)
                      : controller.viewModel.styleProvider.auto(FontType.onPrimary),
                ),
                onPressed: () {
                  if (fieldController.getIsValidEvent.invoke()) {
                    var value = fieldController.getValueEvent.invoke();
                    setValue(value);
                    setState(() {
                      isEditMode = false;
                      fieldValue = value;
                    });
                  }
                },
                onHover: (value) {
                  setState(() {
                    isHoverConfirmBtn = value;
                  });
                },
              ),
              VerticalDivider(
                width: controller.viewModel.config.fontSize.data,
              ),
              OutlinedButton(
                style: OutlinedButton.styleFrom(
                  backgroundColor: controller.viewModel.config.colorScheme.primary,
                  primary: controller.viewModel.config.colorScheme.tertiary,
                  padding: EdgeInsets.all(
                    controller.viewModel.config.fontSize.data,
                  ),
                  side: BorderSide(
                    color:
                        isHoverCancelBtn ? controller.viewModel.config.colorScheme.tertiary : controller.viewModel.config.colorScheme.onPrimary,
                  ),
                  shape: const RoundedRectangleBorder(
                    borderRadius: BorderRadius.zero,
                  ),
                ),
                child: Text(
                  "Cancelar",
                  style: isHoverCancelBtn
                      ? controller.viewModel.styleProvider.auto(FontType.onTertiaryContainer)
                      : controller.viewModel.styleProvider.auto(FontType.onPrimary),
                ),
                onPressed: () {
                  setState(() {
                    isEditMode = false;
                    isHover = false;
                    isHoverCancelBtn = false;
                  });
                  fieldController.resetEvent.invoke();
                },
                onHover: (value) {
                  setState(() {
                    isHoverCancelBtn = value;
                  });
                },
              ),
            ],
          )
        ],
      ),
    );
  }
}

class EditorFrame extends StatefulWidget {
  // ignore: use_key_in_widget_constructors
  const EditorFrame({
    required this.hintText,
    required this.child,
    required this.controller,
    this.margin,
  });

  final String hintText;
  final Widget child;
  final FormControllerOld controller;
  final EdgeInsets? margin;

  @override
  State<EditorFrame> createState() {
    // ignore: no_logic_in_create_state
    return _EditorFrameState(
      hintText: hintText,
      child: child,
      controller: controller,
      margin: margin,
    );
  }
}

class _EditorFrameState extends State<EditorFrame> {
  _EditorFrameState({
    required this.hintText,
    required this.child,
    required this.controller,
    this.margin,
  });

  final String hintText;
  final Widget child;
  final FormControllerOld controller;
  final EdgeInsets? margin;
  // inner state
  bool isHover = false;
  bool isEditMode = false;
  bool isHoverConfirmBtn = false;
  bool isHoverCancelBtn = false;

  @override
  Widget build(BuildContext context) {
    return Container(
      margin: margin,
      child: AnimatedSwitcher(
        duration: controller.viewModel.config.animationDuration,
        child: isEditMode ? editor() : button(),
      ),
    );
  }

  Widget button() {
    return OutlinedButton(
      style: OutlinedButton.styleFrom(
        backgroundColor: controller.viewModel.config.colorScheme.primary,
        padding: const EdgeInsets.all(30),
        primary: controller.viewModel.config.colorScheme.tertiary,
        side: BorderSide(
          color: isHover ? controller.viewModel.config.colorScheme.onTertiaryContainer : controller.viewModel.config.colorScheme.onPrimary,
        ),
        shape: const RoundedRectangleBorder(
          borderRadius: BorderRadius.zero,
        ),
      ),
      child: Row(
        children: [
          Icon(
            Icons.edit,
            color: controller.viewModel.config.colorScheme.tertiary,
            size: controller.viewModel.config.fontSize.data,
          ),
          const VerticalDivider(),
          Flexible(
            child: Row(
              children: [
                Text(
                  hintText,
                  maxLines: 3,
                  softWrap: true,
                  style: controller.viewModel.styleProvider.auto(FontType.onTertiaryContainer),
                ),
              ],
            ),
          ),
        ],
      ),
      onHover: (value) {
        setState(() {
          isHover = value;
        });
      },
      onPressed: () {
        setState(() {
          isEditMode = true;
        });
      },
    );
  }

  Widget editor() {
    return Container(
      padding: const EdgeInsets.all(7),
      decoration: BoxDecoration(
        border: Border.all(
          color: controller.viewModel.config.colorScheme.tertiary,
        ),
      ),
      child: Column(
        children: [
          child,
          Row(
            mainAxisAlignment: MainAxisAlignment.end,
            children: [
              OutlinedButton(
                style: OutlinedButton.styleFrom(
                  backgroundColor: controller.viewModel.config.colorScheme.primary,
                  primary: controller.viewModel.config.colorScheme.tertiary,
                  padding: EdgeInsets.all(
                    controller.viewModel.config.fontSize.data,
                  ),
                  side: BorderSide(
                    color: isHoverConfirmBtn
                        ? controller.viewModel.config.colorScheme.onTertiaryContainer
                        : controller.viewModel.config.colorScheme.onPrimary,
                  ),
                  shape: const RoundedRectangleBorder(
                    borderRadius: BorderRadius.zero,
                  ),
                ),
                child: Text(
                  "Confirmar",
                  style: isHoverConfirmBtn
                      ? controller.viewModel.styleProvider.auto(FontType.onTertiaryContainer)
                      : controller.viewModel.styleProvider.auto(FontType.onPrimary),
                ),
                onPressed: () {
                  if (controller.getIsFormValid()) {
                    setState(() {
                      isEditMode = false;
                    });
                  }
                },
                onHover: (value) {
                  setState(() {
                    isHoverConfirmBtn = value;
                  });
                },
              ),
              VerticalDivider(
                width: controller.viewModel.config.fontSize.data,
              ),
              OutlinedButton(
                style: OutlinedButton.styleFrom(
                  backgroundColor: controller.viewModel.config.colorScheme.primary,
                  primary: controller.viewModel.config.colorScheme.tertiary,
                  padding: EdgeInsets.all(
                    controller.viewModel.config.fontSize.data,
                  ),
                  side: BorderSide(
                    color:
                        isHoverCancelBtn ? controller.viewModel.config.colorScheme.tertiary : controller.viewModel.config.colorScheme.onPrimary,
                  ),
                  shape: const RoundedRectangleBorder(
                    borderRadius: BorderRadius.zero,
                  ),
                ),
                child: Text(
                  "Cancelar",
                  style: isHoverCancelBtn
                      ? controller.viewModel.styleProvider.auto(FontType.onTertiaryContainer)
                      : controller.viewModel.styleProvider.auto(FontType.onPrimary),
                ),
                onPressed: () {
                  setState(() {
                    isEditMode = false;
                    isHover = false;
                    isHoverCancelBtn = false;
                  });
                },
                onHover: (value) {
                  setState(() {
                    isHoverCancelBtn = value;
                  });
                },
              ),
            ],
          )
        ],
      ),
    );
  }
}
