import 'package:badges/badges.dart';
import 'package:flutter/material.dart';
// ignore: import_of_legacy_library_into_null_safe
import 'package:flutter_icons/flutter_icons.dart';
import 'package:modular_system_core/mscore.dart';

class EntityTableAppBar<T> extends StatefulWidget {
  // ignore: use_key_in_widget_constructors
  const EntityTableAppBar({
    required this.viewModel,
    required this.appViewModel,
    required this.updateTableFuture,
  });

  final AppViewModel appViewModel;
  final EntityTableViewModel<T> viewModel;
  final Future updateTableFuture;

  @override
  State<EntityTableAppBar> createState() {
    // ignore: no_logic_in_create_state
    return _EntityTableAppBarState<T>(
      viewModel: appViewModel,
      tableviewModel: viewModel,
    );
  }
}

class _EntityTableAppBarState<T> extends State<EntityTableAppBar> {
  _EntityTableAppBarState({
    required this.viewModel,
    required this.tableviewModel,
  }) {
    subscribeToEvents();
  }

  AppViewModel viewModel;
  EntityTableViewModel<T> tableviewModel;
  bool isSelectorOn = false;
  late bool showSelectorActions = true;
  late int filtersCount = tableviewModel.state.search.filters.where((x) => x.isActive).length;
  late Future updateFuture = tableviewModel.state.updateFuture;

  void subscribeToEvents() {
    tableviewModel.events.update.subscribe(onUpdate);
    tableviewModel.events.searchFiltersChanged.subscribe(onSearchFiltersChanged);
  }

  void unsubscribeToEvents() {
    tableviewModel.events.update.unsubscribe(onUpdate);
    tableviewModel.events.searchFiltersChanged.unsubscribe(onSearchFiltersChanged);
  }

  @override
  void dispose() {
    unsubscribeToEvents();
    super.dispose();
  }

  // !try adding an app bar inside the bottom section of the sliver app bar!
  @override
  Widget build(BuildContext context) {
    return SliverAppBar(
      pinned: true,
      elevation: 10,
      backgroundColor: viewModel.config.colorScheme.background,
      title: TextField(
        controller: tableviewModel.state.searchbarController,
        style: viewModel.styleProvider.auto(FontType.onBackground, FontSize.subtitle),
        decoration: InputDecoration(
          hintText: "Pesquisar ...",
          hintStyle: viewModel.styleProvider.auto(FontType.onBackground, FontSize.subtitle),
          focusedBorder: UnderlineInputBorder(
            borderSide: BorderSide(
              color: viewModel.config.colorScheme.onPrimary,
            ),
          ),
          border: InputBorder.none,
        ),
        onChanged: onTextFieldChanged,
      ),
      leading: IconButton(
        icon: Icon(
          Icons.search,
          color: viewModel.config.colorScheme.onPrimary,
          size: viewModel.config.fontSize.subtitle,
        ),
        onPressed: () {
          tableviewModel.events.update.invoke();
        },
      ),
      actions: [
        IconButton(
          padding: EdgeInsets.symmetric(
            horizontal: viewModel.config.fontSize.subtitle * 0.5,
          ),
          icon: Icon(
            Icons.clear,
            color: viewModel.config.colorScheme.onPrimary,
            size: viewModel.config.fontSize.subtitle,
          ),
          onPressed: () {
            tableviewModel.state.searchbarController.text = "";
            clearTextFilter();
          },
        ),
        IconButton(
          padding: EdgeInsets.symmetric(
            horizontal: viewModel.config.fontSize.subtitle * 0.5,
          ),
          icon: Icon(
            Icons.post_add,
            color: viewModel.config.colorScheme.onPrimary,
            size: viewModel.config.fontSize.subtitle,
          ),
          onPressed: tableviewModel.events.showForm.invoke,
        ),
        IconButton(
          padding: EdgeInsets.symmetric(
            horizontal: viewModel.config.fontSize.subtitle * 0.5,
          ),
          icon: Icon(
            MaterialCommunityIcons.restore,
            color: viewModel.config.colorScheme.onPrimary,
            size: viewModel.config.fontSize.subtitle,
          ),
          onPressed: tableviewModel.events.update.invoke,
        ),
      ],
      bottom: AppBar(
        elevation: 0,
        backgroundColor: viewModel.config.colorScheme.background,
        title: Container(
          alignment: Alignment.center,
          //color: Colors.green,
          child: Row(
            mainAxisAlignment: MainAxisAlignment.spaceBetween,
            crossAxisAlignment: CrossAxisAlignment.center,
            children: [
              Row(
                children: [
                  Badge(
                    position: BadgePosition.topEnd(),
                    badgeColor: viewModel.config.colorScheme.tertiary,
                    badgeContent: Text(
                      filtersCount.toString(),
                      style: viewModel.styleProvider.auto(FontType.onTertiary, FontSize.label),
                    ),
                    child: IconButton(
                      icon: Icon(
                        Icons.filter_list,
                        color: viewModel.config.colorScheme.onBackground,
                        size: viewModel.config.fontSize.subtitle,
                      ),
                      onPressed: () {
                        viewModel.showBottomSheet(TableFilterSheet(
                          viewModel: viewModel,
                          tableViewModel: tableviewModel,
                        ));
                      },
                    ),
                  ),
                  tableviewModel.config.hasAppbarActions
                      ? VerticalDivider(
                          color: viewModel.config.colorScheme.onBackground,
                        )
                      : Container(),
                  tableviewModel.config.hasAppbarActions
                      ? IconButton(
                          icon: Icon(
                            tableviewModel.state.extraOptionsOn ? Icons.expand_less : Icons.expand_more,
                            color: tableviewModel.state.extraOptionsOn
                                ? viewModel.config.colorScheme.tertiary
                                : viewModel.config.colorScheme.onBackground,
                            size: viewModel.config.fontSize.subtitle,
                          ),
                          onPressed: () {
                            setState(() {
                              tableviewModel.state.extraOptionsOn = !tableviewModel.state.extraOptionsOn;
                              tableviewModel.events.showExtraOptionsEvent.invoke();
                            });
                          },
                        )
                      : Container(),
                  VerticalDivider(
                    color: viewModel.config.colorScheme.onBackground,
                  ),
                  IconButton(
                    icon: Icon(
                      Icons.check_box,
                      color: tableviewModel.state.selectorOn ? viewModel.config.colorScheme.tertiary : viewModel.config.colorScheme.onBackground,
                      size: viewModel.config.fontSize.subtitle,
                    ),
                    onPressed: () {
                      setState(() {
                        tableviewModel.events.setSelectorMode.invoke(
                          !tableviewModel.state.selectorOn,
                        );
                      });
                    },
                  ),
                  VerticalDivider(
                    color: viewModel.config.colorScheme.onBackground,
                  ),
                  getSelectorDropdown(),
                  VerticalDivider(
                    color: viewModel.config.colorScheme.onBackground,
                  ),
                  getAppbarActions(),
                  FutureBuilder(
                    future: updateFuture,
                    builder: (context, snapshot) {
                      if (snapshot.connectionState == ConnectionState.done && !snapshot.hasError) {
                        return Row(
                          children: [
                            Padding(
                              padding: const EdgeInsets.symmetric(
                                horizontal: 5,
                              ),
                              child: Icon(
                                MaterialCommunityIcons.table_settings,
                                color: viewModel.config.colorScheme.onBackground,
                                size: viewModel.config.fontSize.subtitle,
                              ),
                            ),
                            Text(
                              "${getPagination()} Total: ${tableviewModel.state.paginationOut.total}",
                              style: viewModel.styleProvider.auto(FontType.onBackground, FontSize.subtitle),
                            )
                          ],
                        );
                      } else if (snapshot.hasError) {
                        return Row(
                          children: [
                            Padding(
                              padding: const EdgeInsets.symmetric(
                                horizontal: 5,
                              ),
                              child: Icon(
                                MaterialCommunityIcons.alert,
                                color: viewModel.config.colorScheme.onBackground,
                                size: viewModel.config.fontSize.subtitle,
                              ),
                            ),
                            Text(
                              "Erro",
                              style: viewModel.styleProvider.auto(FontType.onError, FontSize.subtitle),
                            )
                          ],
                        );
                      } else {
                        return Row(
                          children: [
                            Padding(
                              padding: const EdgeInsets.symmetric(
                                horizontal: 5,
                              ),
                              child: Icon(
                                MaterialCommunityIcons.database_refresh,
                                color: viewModel.config.colorScheme.onBackground,
                                size: viewModel.config.fontSize.subtitle,
                              ),
                            ),
                            const VerticalDivider(),
                            Text(
                              "Carregando...",
                              style: viewModel.styleProvider.auto(FontType.onBackground, FontSize.subtitle),
                            )
                          ],
                        );
                      }
                    },
                  ),
                ],
              ),
              Row(
                mainAxisAlignment: MainAxisAlignment.end,
                children: [
                  FutureBuilder(
                    future: updateFuture,
                    builder: (context, snapshot) {
                      if (snapshot.connectionState == ConnectionState.done && !snapshot.hasError) {
                        return Container(
                          margin: const EdgeInsets.symmetric(
                            horizontal: 25,
                          ),
                          child: Row(
                            children: [
                              IconButton(
                                iconSize: viewModel.config.fontSize.subtitle,
                                mouseCursor: previousPageExists() ? null : SystemMouseCursors.basic,
                                splashRadius: previousPageExists() ? null : 1,
                                color: previousPageExists() ? viewModel.config.colorScheme.tertiary : viewModel.config.colorScheme.onBackground,
                                padding: const EdgeInsets.symmetric(
                                  horizontal: 15,
                                ),
                                icon: const Icon(
                                  Icons.navigate_before,
                                ),
                                onPressed: () {
                                  if (previousPageExists()) {
                                    tableviewModel.state.search.pagination.offset =
                                        tableviewModel.state.search.pagination.offset - tableviewModel.state.search.pagination.limit;
                                    tableviewModel.events.update.invoke();
                                  }
                                },
                              ),
                              IconButton(
                                iconSize: viewModel.config.fontSize.subtitle,
                                mouseCursor: nextPageExists() ? null : SystemMouseCursors.basic,
                                splashRadius: nextPageExists() ? null : 1,
                                color: nextPageExists() ? viewModel.config.colorScheme.tertiary : viewModel.config.colorScheme.onBackground,
                                padding: const EdgeInsets.symmetric(
                                  horizontal: 15,
                                ),
                                icon: const Icon(
                                  Icons.navigate_next,
                                ),
                                onPressed: () {
                                  if (nextPageExists()) {
                                    tableviewModel.state.search.pagination.offset =
                                        tableviewModel.state.search.pagination.offset + tableviewModel.state.search.pagination.limit;
                                    tableviewModel.events.update.invoke();
                                  }
                                },
                              ),
                            ],
                          ),
                        );
                      } else if (snapshot.hasError) {
                        return Text(
                          "ERRO",
                          style: viewModel.styleProvider.auto(FontType.onError, FontSize.subtitle),
                        );
                      } else {
                        return Container();
                      }
                    },
                  ),
                  DropdownButton(
                    value: tableviewModel.state.search.pagination.limit,
                    focusColor: Colors.transparent,
                    dropdownColor: viewModel.config.colorScheme.primary,
                    iconEnabledColor: viewModel.config.colorScheme.onBackground,
                    underline: SizedBox(
                      height: 1.5,
                      child: Container(
                        color: viewModel.config.colorScheme.onBackground,
                      ),
                    ),
                    items: <DropdownMenuItem<int>>[
                      DropdownMenuItem(
                        value: 10,
                        child: Text(
                          "10",
                          style: viewModel.styleProvider.auto(FontType.onBackground, FontSize.subtitle),
                        ),
                      ),
                      DropdownMenuItem(
                        value: 30,
                        child: Text(
                          "30",
                          style: viewModel.styleProvider.auto(FontType.onBackground, FontSize.subtitle),
                        ),
                      ),
                      DropdownMenuItem(
                        value: 50,
                        child: Text(
                          "50",
                          style: viewModel.styleProvider.auto(FontType.onBackground, FontSize.subtitle),
                        ),
                      ),
                      DropdownMenuItem(
                        value: 100,
                        child: Text(
                          "100",
                          style: viewModel.styleProvider.auto(FontType.onBackground, FontSize.subtitle),
                        ),
                      ),
                    ],
                    onChanged: (limit) {
                      setState(() {
                        tableviewModel.state.search.pagination.limit = limit as int;
                        tableviewModel.events.update.invoke();
                      });
                    },
                  ),
                ],
              ),
            ],
          ),
        ),
      ),
    );
  }

  Widget getSelectorDropdown() {
    if (!showSelectorActions) {
      return Container();
    }

    return PopupMenuButton(
      icon: Icon(
        Icons.more_vert,
        color: viewModel.config.colorScheme.onBackground,
        size: viewModel.config.fontSize.subtitle,
      ),
      itemBuilder: (context) {
        var items = <PopupMenuItem<int>>[];
        var actions = tableviewModel.properties.selectorActionsBuilder!(tableviewModel);
        int index;
        for (index = 0; index < actions.length; index++) {
          var action = actions.elementAt(index);
          var item = PopupMenuItem(
            value: index,
            child: action.key,
            onTap: () => action.value.call(),
          );
          items.add(item);
        }
        items.add(PopupMenuItem(
          value: index,
          child: Row(
            children: [
              Icon(
                Icons.cancel,
                color: viewModel.config.colorScheme.onBackground,
                size: viewModel.config.fontSize.data,
              ),
              const VerticalDivider(),
              Text(
                'Cancelar',
                style: viewModel.styleProvider.auto(FontType.onBackground),
              ),
            ],
          ),
          onTap: () {},
        ));
        return items;
      },
    );
  }

  String getPagination() {
    var pagination = tableviewModel.state.paginationOut;
    var currentPage = pagination.offset / pagination.limit + 1;
    var totalPages = pagination.total / pagination.limit + 1;
    return "$currentPage de ${totalPages.floor()}";
  }

  bool nextPageExists() {
    return tableviewModel.state.paginationOut.total > (tableviewModel.state.paginationOut.offset + tableviewModel.state.paginationOut.limit);
  }

  bool previousPageExists() {
    return tableviewModel.state.paginationOut.offset > 0;
  }

  Widget getAppbarActions() {
    if (tableviewModel.properties.appbarActions != null && tableviewModel.properties.appbarActions!.isNotEmpty) {
      var children = <Widget>[];
      for (var item in tableviewModel.properties.appbarActions!) {
        children.add(
          IconButton(
            icon: Icon(
              item.key,
              color: viewModel.config.colorScheme.onBackground,
              size: viewModel.config.fontSize.subtitle,
            ),
            onPressed: () {
              item.value.call(tableviewModel);
            },
          ),
        );
      }
      children.add(const VerticalDivider());
      return Row(
        children: children,
      );
    }
    return Container();
  }

  void onTextFieldChanged(String value) {
    tableviewModel.events.setSearchText.invoke(value);
  }

  void onUpdate() {
    setState(() {
      updateFuture = tableviewModel.state.updateFuture;
    });
  }

  void onSearchFiltersChanged() {
    setState(() {
      filtersCount = tableviewModel.state.search.filters.where((x) => x.isActive).length;
    });
  }

  void clearTextFilter() {
    tableviewModel.events.setSearchText.invoke("");
  }
}
